import codecs
import hashlib
import logging
import sys
import json
import random

# Optionally, we can provide ChaCha20Poly1305 to support securing the channel.  Required if the
# --en/decrypt option is used.
try:
    from chacha20poly1305 import ChaCha20Poly1305
except ImportError:
    pass

from ..types		import Account

log				= logging.getLogger( __package__ )


def chacha20poly1305( password: str ) -> ChaCha20Poly1305:
    """Stretch the password (w/ sha256 without a salt!, not pbkdf2_hmac, as this must be easily
    replicable in the receiver who has only the password).  The security requirements of this
    channel is not high, because only public wallet addresses are being transported, so this is
    acceptable.  A one-time Nonce must be used whenever the same password is used, though.

    """
    key				= hashlib.sha256()
    key.update( password.encode( 'UTF-8' ))
    return ChaCha20Poly1305( key=key.digest() )


def nonce_add(
    nonce: bytes,
    offset: int,
):
    nonce_len			= len( nonce )
    nonce_int			= int.from_bytes( nonce, 'big' )
    nonce_now			= (( nonce_int + int( offset )) % 2**( 8 * nonce_len ))
    return nonce_now.to_bytes( nonce_len, 'big' )


def accountgroups_input(
    cipher	= None,		# Are input accountgroups records encrypted?
    file	= None,		# Where to retrieve input from
    encoding	= None,		# Does channel require decoding from binary? Use this encoding, if so
    healthy	= None,		# Is file healthy for reading?  Read and ignore input while not.
):
    """Receive and yield accountgroups, ignoring any that cannot be parsed, or received while not
    healthy.  Add the enumeration to the nonce for decrypting.

    The session nonce must be recovered from the first line of input.

    """
    if file is None:
        file			= sys.stdin
    nonce			= None
    while True:
        # Attempt to receive a record, while connection is healthy
        try:
            record		= None
            while not record or not record.endswith( b'\n' if type(record) is bytes else '\n' ):
                recv		= file.readline()
                health		= healthy is None or healthy( file )
                if not health:
                    if recv:
                        log.warning( f"{file!r:.32} Unhealthy; ignoring input: {recv!r}" )
                    break
                if recv:
                    if record is None:
                        record	= recv
                    else:
                        record += recv
        except EOFError:
            # Session has terminated; TODO: yield the EOFError to signal no more inputs (ever) available?
            return
        if not health:
            yield None,None
            continue

        # Got a record on a healthy connection!
        if encoding:  # Eg. if file is binary (eg. a Serial device), decode
            try:
                record		= record.decode( encoding )
            except Exception as exc:
                log.warning( f"Discarding invalid record {record!r}: {exc!r}" )
                yield None, None
                continue

        # Ignore empty records
        record			= record.strip()
        if not record:
            continue

        # See if records are indexed.  Only int or 'nonce' is accepted for index.
        index			= None
        try:
            index,payload	= record.split( ':', 1 )        # ValueError if not <index>:<payload>
            index		= int( index )			# ValueError if <index> not int
        except ValueError:
            if index != 'nonce':				# Otherwise, only 'nonce': <payload> acceptable
                index,payload	= None,record			# ...if not; then records are not indexed.

        try:
            if cipher:
                if nonce is None or index == 'nonce':
                    # Either this is the first record ever received, *or* the counterparty has
                    # restarted and/or is re-noncing the session.
                    try:
                        assert index == 'nonce', \
                            f"Failed to find 'nonce' enumeration prefix on first record: {record!r}"
                        ciphertext	= bytearray( codecs.decode( payload.strip(), 'hex_codec' ))
                        nonce		= bytes( cipher.decrypt( b'\x00' * 12, ciphertext ))
                    except Exception as exc:
                        message		= f"Failed to recover nonce from {record!r}; cannot proceed: {exc}"
                        log.error( message )
                        return message
                    log.info( f"Decrypting accountgroups with nonce: {nonce.hex()}" )
                    continue

                nonce_now	= nonce_add( nonce, index )
                ciphertext	= bytearray( bytes.fromhex( payload ))
                plaintext 	= bytes( cipher.decrypt( nonce_now, ciphertext ))
                payload		= plaintext.decode( 'UTF-8' )
            group		= json.loads( payload )
        except Exception as exc:
            log.warning( f"Discarding invalid record {record!r}: {exc!r}" )
            yield None, None
        else:
            yield int(index), group


def file_outputline(
    file,
    output,
    encoding	= None,
    flush	= True,
    healthy	= None,
):
    """Returns the health of the file at the end of the output write/flush.  Unhealthy file_output
    should return False, allowing the caller to re-try opening the channel and outputting the
    record.

    A predicate 'healthy' takes a file that tests for its health, returns True iff healhty, False or raising
    Exception if not healthy.

    """
    if file is None:
        file			= sys.stdout

    output		       += '\n'
    if encoding:
        output			= output.encode( encoding )

    # Confirm that the file is healthy before writing output
    if ( health := healthy is None or healthy( file )):
        log.info( f"File {file!r:.36} writing {len(output):3}: {output!r:.36}{'...' if len(output) >36 else ''}" )
        file.write( output )
    if not health:
        log.warning( f"File {file!r:.36} became unhealthy before output of {output!r}" )
        return health

    # Confirm that the file was healthy right up 'til the buffer is done flushing
    while ( health := healthy is None or healthy( file )) and flush:
        '''
        if flush and hasattr( file, 'out_waiting' ):
            if ( waiting := file.out_waiting ) > 0:
                log.info( f"{file!r:.32} waiting for {waiting} bytes to flush" )
                time.sleep( 1/100 )
                continue
            flush		= False
        elif flush:
            file.flush()
            flush		= False
        '''
        file.flush()
        flush			= False

    if not health:
        log.warning( f"File {file!r:.36} became unhealthy during output of {output!r}" )
    return health


def accountgroups_output(
    group,
    index	= None,
    cipher	= None,
    nonce	= None,
    file	= None,
    flush	= True,
    corrupt	= 0,
    nonce_emit	= True,		# force encrypted Nonce to be emitted
    encoding	= None,		# Does channel require encoding to binary? Use this encoding, if so
    healthy	= None,		# Detect health of file
):
    """Emit accountgroup records to the provided file, or sys.stdout.

    For each record, we will support either a sequence of Accounts (produced by accountgroups()), or
    a sequence of tuples of (<crypto>, <path>, <address>), (ie. recovered via accountgroups_input.

    Supports binary file-like objects (eg pyserial.Serial) w/ the encoding parameter.  Outputs one line,
    blocking forever -- the counterparty can (and likely will) block for an indeterminate amount of time,
    as it waits to use the provided account groups.

    Returns the health of the output before, during and after attempted output write/flush.

    """
    assert not cipher or ( nonce and index is not None ), \
        "Encryption requires both nonce and index"

    # For encrypted output, the first record emitted always must be the encrypted nonce.  This might
    # occur multiple times outputting a stream of data, if it is detected that the counterparty has
    # restarted.  Serial port pins such as DTR/DSR can be used to detect counterparty disconnection.
    # Always emit extra leading newlines to restart the counterparty line reading, since we don't
    # know what kind of input remains in its buffer.
    if cipher and nonce and nonce_emit:
        # Emit the one-time record containing the encrypted nonce, itself w/ a zero nonce.
        plaintext		= bytearray( nonce )
        ciphertext		= bytes( cipher.encrypt( b'\x00' * len( nonce ), plaintext ))
        record			= ( 'nonce', ciphertext.hex(), )
        output			= "\n\n" + ": ".join( record )
        log.info( f"Encrypting accountgroups with nonce: {nonce.hex()}" )
        if not file_outputline( file, output, encoding=encoding, flush=flush, healthy=healthy ):
            return False

    if not group or ( cipher and index is None ):
        return		# Ignore un-parsable/empty groups, or encrypting w/ no index

    # Emit the (optionally encrypted and indexed) accountgroup record.
    payload			= json.dumps([
        (acct.crypto, acct.path, acct.address) if isinstance( acct, Account ) else acct
        for acct in group
    ])
    if cipher:
        plaintext		= bytearray( payload.encode( 'UTF-8' ))
        nonce_now		= nonce_add( nonce, index )
        ciphertext		= bytes( cipher.encrypt( nonce_now, plaintext ))
        record			= ( codecs.encode( ciphertext, 'hex_codec' ).decode( 'ascii' ), )
    else:
        record			= ( payload, )

    if index is not None:
        record			= ( f"{index:>5}", ) + record

    output			= ": ".join( record )
    if corrupt:
        fraction		= corrupt / 100
        output			= ''.join(
            random.choice( 'abcdefghijklmnopqrstuvwxyz0123456789' ) if random.random() < fraction else c
            for c in output
        )

    # Finally, output the record, returning the health of the file at the end of the transmission
    return file_outputline( file, output, encoding=encoding, flush=flush, healthy=healthy )



import argparse
import codecs
import hashlib
import logging
import math
import os

import PySimpleGUI as sg

from ..types		import Account
from ..api		import create, group_parser, random_secret
from ..recovery		import recover, recover_bip39
from ..util		import log_level, log_cfg
from ..layout		import write_pdfs
from ..defaults		import GROUPS, GROUP_THRESHOLD_RATIO, CRYPTO_PATHS, CARD, CARD_SIZES

log				= logging.getLogger( __package__ )

font				= ('Courier', 14)
font_small			= ('Courier', 11)

I_kwds				= dict(
    change_submits	= True,
    font		= font,
)
T_kwds				= dict(
    font		= font,
)
F_kwds				= dict(
    font		= font,
)
B_kwds				= dict(
    font		= font,
    enable_events	= True,
)


def groups_layout( names, group_threshold, groups, passphrase=None ):
    """Return a layout for the specified number of SLIP-39 groups.

    """

    group_body			= [
        sg.Frame(
            '#', [[ sg.Column( [
                [ sg.Text( f'{g+1}', **T_kwds ) ]
                for g,(g_name,(g_need,g_size)) in enumerate( groups.items() )
            ], key='-GROUP-NUMBER-' ) ]], **F_kwds
        ),
        sg.Frame(
            'Recovery Group', [[ sg.Column( [
                [ sg.Input( f'{g_name}', key=f'-G-NAME-{g}', **I_kwds ) ]
                for g,(g_name,(g_need,g_size)) in enumerate( groups.items() )
            ], key='-GROUP-NAMES-' ) ]], **F_kwds
        ),
        sg.Frame(
            'Needs at least', [[ sg.Column( [
                [ sg.Input( f'{g_need}', key=f'-G-NEED-{g}', **I_kwds ) ]
                for g,(g_name,(g_need,g_size)) in enumerate( groups.items() )
            ], key='-GROUP-NEEDS-' ) ]], **F_kwds
        ),
        sg.Frame(
            'of Cards in Group', [[ sg.Column( [
                [ sg.Input( f'{g_size}', key=f'-G-SIZE-{g}', **I_kwds ) ]
                for g,(g_name,(g_need,g_size)) in enumerate( groups.items() )
            ], key='-GROUP-SIZES-' ) ]], **F_kwds
        ),
    ]
    prefix			= (32, 1)
    inputs			= (32, 1)
    layout                      = [
        [
            sg.Frame( 'Output PDF File(s)', [
                [
                    sg.Text( "Save PDF(s) to (ie. USB drive): ",                size=prefix,    **T_kwds ),
                    sg.Input( sg.user_settings_get_entry( "-target folder-", ""),
                                                        key='-TARGET-',         size=inputs,    **I_kwds ),  # noqa: E127
                    sg.FolderBrowse( **B_kwds ),
                ],
                [
                    sg.Text( "Seed Name(s): ",                                  size=prefix,    **T_kwds ),
                    sg.Input( f"{', '.join( names )}",  key='-NAMES-',          size=inputs,    **I_kwds ),
                    sg.Text( "(default is 'SLIP39...'; comma-separated)",                       **T_kwds ),
                ],
                [
                    sg.Text( "Card size: ",                                     size=prefix,    **T_kwds ),
                ] + [
                    sg.Radio( f"{card}", "CS",          key=f"-CS-{card}", default=(card == CARD), **B_kwds )
                    for card in CARD_SIZES
                ],
            ],                                                  key='-OUTPUT-F-',               **F_kwds ),
        ],
    ] + [
        [
            sg.Frame( 'Seed Data Source (256-bit seeds produce more Mnemonic words to type into your Trezor; 512-bit seeds aren\'t Trezor compatible)', [
                [
                    sg.Radio( "128-bit Random", "SD",   key='-SD-128-RND-',     default=True,   **B_kwds ),
                    sg.Radio( "256-bit Random", "SD",   key='-SD-256-RND-',     default=False,  **B_kwds ),
                    sg.Radio( "512-bit Random", "SD",   key='-SD-512-RND-',     default=False,  **B_kwds ),
                ],
                [
                    sg.Radio( "128-bit Fixed ", "SD",   key='-SD-128-FIX-',     default=False,  **B_kwds ),
                    sg.Radio( "256-bit Fixed ", "SD",   key='-SD-256-FIX-',     default=False,  **B_kwds ),
                    sg.Radio( "512-bit Fixed ", "SD",   key='-SD-512-FIX-',     default=False,  **B_kwds ),
                ],
                [
                    sg.Radio( "512-bit BIP-39", "SD",   key='-SD-BIP-',         default=False,  **B_kwds ),
                    sg.Radio( "SLIP-39",        "SD",   key='-SD-SLIP-',        default=False,  **B_kwds ),
                ],
                [
                    sg.Frame( 'From', [
                        [
                            sg.Text( "Mnemonic(s): ",   key='-SD-DATA-T-',      size=prefix,    **T_kwds ),
                            sg.Multiline( "",           key='-SD-DATA-',        size=(128,1),   **I_kwds ),
                        ],
                    ],                                  key='-SD-DATA-F-',      visible=False,  **F_kwds ),
                ],
                [
                    sg.Frame( 'Passphrase (if provided when Mnemonic was created)', [
                        [
                            sg.Text( "Passphrase (decrypt): ",                  size=prefix,    **T_kwds ),
                            sg.Input( "",               key='-SD-PASS-',        size=inputs,    **I_kwds ),
                        ],
                    ],                                  key='-SD-PASS-F-',      visible=False,  **F_kwds ),
                ],
                [
                    sg.Text( "Seed Data: ",                                     size=prefix,    **T_kwds ),
                    sg.Text( "",                        key='-SD-SEED-',        size=(128,1),   **T_kwds ),
                ],
            ],                                          key='-SD-SEED-F-',                      **F_kwds ),
        ],
    ] + [
        [
            sg.Frame( '⊻ XOR Extra Seed Entropy (eg. Die rolls, ...)', [
                [
                    sg.Radio( "Hex",              "SE", key='-SE-HEX-',         default=True,   **B_kwds ),
                    sg.Radio( "SHA-512 Stretched","SE", key='-SE-SHA-',         default=False,  **B_kwds ),
                ],
                [
                    sg.Text( "Entropy (Hex): ",         key='-SE-DATA-T-',      size=prefix,    **T_kwds ),
                    sg.Input( "",                       key='-SE-DATA-',        size=(128,1),   **I_kwds ),
                ],
                [
                    sg.Text( "Seed Entropy: ",                                  size=prefix,    **T_kwds ),
                    sg.Text( "",                        key='-SE-SEED-',        size=(128,1),   **T_kwds ),
                ],
            ],                                          key='-SE-SEED-F-',                      **F_kwds ),
        ]
    ] + [
        [
            sg.Frame( 'Master Secret', [
                [
                    sg.Text( "Seed: ",                                          size=prefix,    **T_kwds ),
                    sg.Text( "",                        key='-SEED-',           size=(128,1),   **T_kwds ),
                ],
            ],                                          key='-SEED-F-',                         **F_kwds ),
        ],
    ] + [
        [
            sg.Frame( 'Groups', [
                [
                    sg.Column( [
                        [
                            sg.Text( "Requires recovery of at least: ",         size=prefix,    **T_kwds ),
                            sg.Input( f"{group_threshold}", key='-THRESHOLD-',  size=inputs,    **I_kwds ),
                            sg.Text( f"(of {len(groups)} SLIP-39 Recovery Groups)",
                                                        key='-RECOVERY-',                       **T_kwds ),  # noqa: E127
                        ],
                        [
                            sg.Text( "Passphrase (encrypt): ",                  size=prefix,    **T_kwds ),
                            sg.Input( f"{passphrase or ''}",
                                                        key='-PASSPHRASE-',     size=inputs,    **I_kwds ),  # noqa: E127
                            sg.Text( "(NOT Trezor compatible, and must be saved separately!!)", **T_kwds ),
                        ],
                        [
                            sg.Button( '+', **B_kwds ),
                        ],
                        group_body,
                    ] ),
                ],
            ],                                          key='-GROUPS-F-',                       **F_kwds ),
        ],
    ] + [
        [
            sg.Frame( 'Summary', [
                [
                    sg.Text(                            key='-SUMMARY-',                        **T_kwds ),
                ]
            ],                                          key='-SUMMARY-F-',                      **F_kwds ),
        ],
        [
            sg.Button( 'Save', **B_kwds ), sg.Button( 'Exit', **B_kwds ),
        ],
        [
            sg.Frame( 'Status', [
                [
                    sg.Text(                            key='-STATUS-',                         **T_kwds ),
                ]
            ],                                          key='-STATUS-F-',                       **F_kwds ),
        ],
    ] + [
        [
            sg.Frame( 'SLIP39 Mnemonics Output', [
                [
                    sg.Multiline( "",                   key='-MNEMONICS-',      size=(190,10),   font=font_small )
                ]
            ], **F_kwds ),
        ],
    ] + [
        [
            sg.Frame( 'Seed Recovery from SLIP39 Mnemonics', [
                [
                    sg.Text( "Seed Verified: ",                                 size=prefix,    **T_kwds ),
                    sg.Text( "",                        key='-SEED-RECOVERED-', size=(128,1),   **T_kwds ),
                ],
            ],                                          key='-RECOVERED-F-',                    **F_kwds ),
        ],
    ]
    return layout


def update_seed_data( window, values ):
    """Respond to changes in the desired Seed Data source, and recover/generate and update the
    -SD-SEED-.  Stores the last known state of the -SD-... radio buttons, and saves/restores the
    user data being supplied on for BIP/SLIP/FIX.

    """
    changed			= False
    dat,pwd			= values['-SD-DATA-'],values['-SD-PASS-']
    try:
        master_secret		= codecs.decode( values['-SD-SEED-'], 'hex_codec' )
    except Exception:
        master_secret		= b''
    for src in [
            '-SD-128-RND-',
            '-SD-256-RND-',
            '-SD-512-RND-',
            '-SD-128-FIX-',
            '-SD-256-FIX-',
            '-SD-512-FIX-',
            '-SD-BIP-',
            '-SD-SLIP-',
    ]:
        # See what we got for -SD-DATA-, for this -SD-... radio button selection
        if values[src] and update_seed_data.src != src:
            # If selected radio-button for Seed Data source changed, save last source's working data
            # and restore what was, last time we were working on this source.
            if update_seed_data.src:
                update_seed_data.was[update_seed_data.src] = dat,pwd
            changed		= True
            update_seed_data.src = src
            dat,pwd		= update_seed_data.was.get( src, ('','') )
            master_secret	= b''
            window['-SD-DATA-'].update( dat )
            window['-SD-PASS-'].update( pwd )
            # And change visibility of Seed Data source controls
            if 'FIX' in update_seed_data.src:
                window['-SD-DATA-T-'].update( "Hex data: " )
                window['-SD-DATA-F-'].update( visible=True  )
                window['-SD-PASS-F-'].update( visible=False )
            elif 'BIP' in update_seed_data.src:
                window['-SD-DATA-T-'].update( "BIP-39 Mnemonic: " )
                window['-SD-DATA-F-'].update( visible=True )
                window['-SD-PASS-F-'].update( visible=True )
            elif 'SLIP' in update_seed_data.src:
                window['-SD-DATA-T-'].update( "SLIP-39 Mnemonics: " )
                window['-SD-DATA-F-'].update( visible=True )
                window['-SD-PASS-F-'].update( visible=True )
            elif 'RND' in update_seed_data.src:
                window['-SD-DATA-F-'].update( visible=False )
                window['-SD-PASS-F-'].update( visible=False )

    # Now that we got our working -SD-DATA- (maybe from last time 'round), compute seed
    if 'BIP' in update_seed_data.src:
        try:
            master_secret	= recover_bip39(
                mnemonic	= dat.strip(),
                passphrase	= pwd.strip().encode( 'UTF-8' )
            )
        except Exception as exc:
            log.exception( f"BIP-39 recovery failed w/ {dat!r} ({pwd!r}): {exc}" )
            return f"Invalid BIP-39 recovery mnemonic: {exc}"
    elif 'SLIP' in update_seed_data.src:
        try:
            master_secret	= recover(
                mnemonics	= dat.strip().split( '\n' ),
                passphrase	= pwd.strip().encode( 'UTF-8' )
            )
        except Exception as exc:
            log.exception( f"SLIP-39 recovery failed w/ {dat!r} ({pwd!r}): {exc}" )
            return f"Invalid SLIP-39 recovery mnemonics: {exc}"
    elif 'FIX' in update_seed_data.src:
        bits			= int( update_seed_data.src.split( '-' )[2] )
        try:
            # 0-fill and truncate any supplied hex data to the desired bit length
            data		= f"{dat:<0{bits//4}.{bits//4}}"
            master_secret 	= codecs.decode( data, 'hex_codec' )
        except Exception as exc:
            return f"Invalid Hex for {bits}-bit fixed seed: {exc}"
    elif changed or not master_secret:  # Random.  Regenerated each time changed, or not valid
        bits			= int( update_seed_data.src.split( '-' )[2] )
        master_secret		= random_secret( bits // 8 )
    # Compute the Seed Data as hex.  Will be 128-, 256- or 512-bit hex data.
    window['-SD-SEED-'].update( codecs.encode( master_secret, 'hex_codec' ).decode( 'ascii' ))
update_seed_data.src		= None  # noqa: E305
update_seed_data.was		= {
    '-SD-BIP-': ("zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong","")  # "<mnemnonic>","<passphrase>"
}


def update_seed_entropy( window, values ):
    """Respond to changes in the extra Seed Entropy, and recover/generate the -SE-SEED-.  It is
    expected to be exactly the same size as the -SD-SEED- data.  Stores the last known state of the
    -SE-... radio buttons, updating visibilties on change.

    """
    for seed_entr_source in [
        '-SE-HEX-',
        '-SE-SHA-',
    ]:
        if values.get( seed_entr_source ) and update_seed_entropy.seed_entr != seed_entr_source:
            update_seed_entropy.seed_entr = seed_entr_source
            # Changed Seed Entropy source!
            if 'HEX' in update_seed_entropy.seed_entr:
                window['-SE-DATA-T-'].update( "Entropy (Hex): " )
            else:
                window['-SE-DATA-T-'].update( "Entropy (Die rolls, etc.): " )

    bits			= len( window['-SD-SEED-'].get() ) * 4
    if 'HEX' in update_seed_entropy.seed_entr:
        try:
            # 0-fill and truncate any supplied hex data to the desired bit length
            data		= f"{values['-SE-DATA-']:<0{bits//4}.{bits//4}}"
            master_entropy	= codecs.decode( data, 'hex_codec' )
        except Exception as exc:
            return f"Invalid Hex for {bits}-bit extra seed entropy: {exc}"
    else:
        try:
            # SHA-512 stretch and possibly truncate
            stretch		= hashlib.sha512()
            stretch.update( values['-SE-DATA-'].encode( 'UTF-8' ))
            master_entropy	= stretch.digest()[:bits//8]
        except Exception as exc:
            return f"Invalid data for {bits}-bit extra seed entropy: {exc}"

    # Compute the Seed Entropy as hex.  Will be 128-, 256- or 512-bit hex data.
    window['-SE-SEED-'].update( codecs.encode( master_entropy, 'hex_codec' ).decode( 'ascii' ))
update_seed_entropy.seed_entr	= None  # noqa: E305


def update_seed_recovered( window, values, details, passphrase=None ):
    """Display the SLIP39 Mnemonics.  Each mnemonic word is maximum 8 characters in length, separated
    by a single space.  The bit- and word-length of each standard SLIP39 Mnemonic is:

        bits words
        ---- -----
        128  20
        256  33
        512  59

    There must always be a passphrase; it may be empty (b'').

    """
    mnemonics			= []
    rows			= []
    for g,(g_of,g_mnems) in details.groups.items() if details else []:
        for i,mnem in enumerate( g_mnems, start=1 ):
            mnemonics.append( mnem )
            mset		= mnem.split()
            while mrow := mset[:20]:
                mset		= mset[20:]
                rows.append( f"{g:<8.8}" + f"{i:2} " + ' '.join( f"{m:<8}" for m in mrow ))
                g		= ''
                i		= ''

    window['-MNEMONICS-'].update( '\n'.join( rows ))

    recohex			= ''
    if mnemonics:
        reco			= recover( mnemonics, passphrase=passphrase or b'' )
        recohex			= codecs.encode( reco, 'hex_codec' ).decode( 'ascii' )
    window['-SEED-RECOVERED-'].update( recohex )


def app(
    names			= None,
    group			= None,
    threshold			= None,
    cryptocurrency		= None,
    passphrase			= None,
):
    """Convert sequence of group specifications into standard { "<group>": (<needs>, <size>) ... }"""

    if isinstance( group, dict ):
        groups			= group
    else:
        groups			= dict(
            group_parser( g )
            for g in group or GROUPS
        )
    assert groups and all( isinstance( k, str ) and len( v ) == 2 for k,v in groups.items() ), \
        f"Each group member specification must be a '<group>': (<needs>, <size>) pair, not {type(next(groups.items()))}"

    group_threshold		= int( threshold ) if threshold else math.ceil( len( groups ) * GROUP_THRESHOLD_RATIO )

    cryptopaths			= []
    for crypto in cryptocurrency or CRYPTO_PATHS:
        try:
            if isinstance( crypto, str ):
                crypto,paths	= crypto.split( ':' )   # Find the  <crypto>,<path> tuple
            else:
                crypto,paths	= crypto		# Already a <crypto>,<path> tuple?
        except ValueError:
            crypto,paths	= crypto,None
        cryptopaths.append( (crypto,paths) )

    sg.theme( 'DarkAmber' )

    sg.user_settings_set_entry( '-target folder-', os.getcwd() )

    layout			= groups_layout( names, group_threshold, groups, passphrase )
    window			= None
    status			= None
    event			= False
    events_termination		= (sg.WIN_CLOSED, 'Exit',)
    master_secret		= None		# default to produce randomly
    while event not in events_termination:
        # Create window (for initial window.read()), or update status
        if window:
            window['-STATUS-'].update( status or 'OK' )
            window['-RECOVERY-'].update( f"(of {len(groups)} SLIP-39 Recovery Groups)", **T_kwds ),
            window['-SD-SEED-F-'].expand( expand_x=True )
            window['-SE-SEED-F-'].expand( expand_x=True )
            window['-SEED-F-'].expand( expand_x=True )
            window['-OUTPUT-F-'].expand( expand_x=True )
            window['-SUMMARY-F-'].expand( expand_x=True )
            window['-STATUS-F-'].expand( expand_x=True )
            window['-RECOVERED-F-'].expand( expand_x=True )
            window['-GROUPS-F-'].expand( expand_x=True )
            timeout		= None 		# Subsequently, block indefinitely
        else:
            window		= sg.Window( f"{', '.join( names or [ 'SLIP39' ] )} Mnemonic Cards", layout )
            timeout		= 0 		# First time through, refresh immediately

        status			= None
        event, values		= window.read( timeout=timeout )
        logging.info( f"{event}, {values}" )
        if not values or event in events_termination:
            continue

        if event == '+' and len( groups ) < 16:
            # Add a SLIP39 Groups row (if not already at limit)
            g			= len( groups )
            name		= f"Group{g+1}"
            needs		= (2,3)
            groups[name] 	= needs
            values[f"-G-NAME-{g}"] = name
            values[f"-G-NEED-{g}"] = needs[0]
            values[f"-G-SIZE-{g}"] = needs[1]
            window.extend_layout( window['-GROUP-NUMBER-'], [[ sg.Text(  f"{g+1}",                          **T_kwds ) ]] )  # noqa: 241
            window.extend_layout( window['-GROUP-NAMES-'],  [[ sg.Input( f"{name}",     key=f"-G-NAME-{g}", **I_kwds ) ]] )  # noqa: 241
            window.extend_layout( window['-GROUP-NEEDS-'],  [[ sg.Input( f"{needs[0]}", key=f"-G-NEED-{g}", **I_kwds ) ]] )  # noqa: 241
            window.extend_layout( window['-GROUP-SIZES-'],  [[ sg.Input( f"{needs[1]}", key=f"-G-SIZE-{g}", **I_kwds ) ]] )  # noqa: 241

        if status := update_seed_data( window, values ):
            update_seed_recovered( window, values, None )
            continue
        if status := update_seed_entropy( window, values ):
            update_seed_recovered( window, values, None )
            continue

        # Compute the Master Secret Seed, from the supplied Seed Data and any extra Seed Entropy
        data			= codecs.decode( window['-SD-SEED-'].get(), 'hex_codec' )
        entr			= codecs.decode( window['-SE-SEED-'].get(), 'hex_codec' )
        seed			= bytes( d ^ e for d,e in zip( data, entr ) )
        window['-SEED-'].update( codecs.encode( seed, 'hex_codec' ).decode( 'ascii' ))

        # A target directory must be selected; use it.  This is where any output will be written.
        # It should usually be a removable volume, but we do not check for this.
        try:
            os.chdir( values['-TARGET-'] )
        except Exception as exc:
            status		= f"Error changing to target directory {values['-TARGET-']!r}: {exc}"
            logging.exception( f"{status}" )
            update_seed_recovered( window, values, None )
            continue

        # Collect up the specified Group names; ignores groups with an empty name (effectively
        # eliminating that group)
        g_rec			= {}
        status			= None
        for g in range( 16 ):
            grp_idx		= f"-G-NAME-{g}"
            if grp_idx not in values:
                break
            grp			= '?'
            try:
                grp		= str( values[grp_idx] ).strip()
                if not grp:
                    continue
                assert grp not in g_rec, \
                    f"Duplicate group name {grp}"
                g_rec[grp] 	= int( values[f"-G-NEED-{g}"] or 0 ), int( values[f"-G-SIZE-{g}"] or 0 )
            except Exception as exc:
                status		= f"Error defining group {g+1} {grp!r}: {exc}"
                logging.exception( f"{status}" )
                update_seed_recovered( window, values, None )
                continue

        # Confirm the selected Group Threshold requirement
        try:
            g_thr_val		= values['-THRESHOLD-']
            g_thr		= int( g_thr_val )
            assert 0 < g_thr <= len( g_rec ), \
                f"must be an integer between 1 and the number of groups ({len(g_rec)})"
        except Exception as exc:
            status		= f"Error defining group threshold {g_thr_val!r}: {exc}"
            logging.exception( f"{status}" )
            update_seed_recovered( window, values, None )
            continue

        # Produce a summary of the recovered SLIP39 Groups, including any passphrase needed for decryption
        summary			= f"Require {g_thr}/{len(g_rec)} Groups, from: {f', '.join( f'{n}({need}/{size})' for n,(need,size) in g_rec.items())}"
        passphrase		= values['-PASSPHRASE-'].strip()
        if passphrase:
            summary	       += f", decrypted w/ passphrase {passphrase!r}"
        passphrase		= passphrase.encode( 'utf-8' ) if passphrase else b''

        window['-SUMMARY-'].update( summary )

        # Deduce the desired Seed names, defaulting to "SLIP39"
        names			= [
            name.strip()
            for name in ( values['-NAMES-'].strip() or "SLIP39" ).split( ',' )
            if name and name.strip()
        ]

        # Compute the SLIP39 Seed details.  For multiple names, each subsequent slip39.create uses a
        # master_secret produced by hashing the prior master_secret entropy.
        details			= {}
        try:
            master_secret	= codecs.decode( window['-SEED-'].get(), 'hex_codec' )
            for name in names:
                log.info( f"SLIP39 for {name} from master_secret: {codecs.encode( master_secret, 'hex_codec' ).decode( 'ascii' )}" )
                details[name]	= create(
                    name		= name,
                    group_threshold	= group_threshold,
                    master_secret	= master_secret,
                    groups		= g_rec,
                    cryptopaths		= cryptopaths,
                    passphrase		= passphrase,
                )
                master_secret	= hashlib.sha512( master_secret ).digest()[:len(master_secret)]
        except Exception as exc:
            status		= f"Error creating: {exc}"
            logging.exception( f"{status}" )
            update_seed_recovered( window, values, None )
            continue

        if status := update_seed_recovered( window, values, details[names[0]], passphrase=passphrase ):
            continue

        # Finally, if all has gone well -- display the resultant <name>/<filename>, and some derived account details
        name_len		= max( len( name ) for name in details )
        status			= '\n'.join(
            f"{name:>{name_len}} == " + ', '.join( f'{a.crypto} @ {a.path}: {a.address}' for a in details[name].accounts[0] )
            for name in details
        )

        # If we get here, no failure status has been detected, and SLIP39 mnemonic and account
        # details { "name": <details> } have been created; we can now save the PDFs; converted
        # details is now { "<filename>": <details> })
        if event == 'Save':
            try:
                card		= next( c for c in CARD_SIZES if values[f"-CS-{c}"] )
                details		= write_pdfs(
                    names	= details,
                    card	= card,	
                )
            except Exception as exc:
                status		= f"Error saving PDF(s): {exc}"
                logging.exception( f"{status}" )
                continue
            name_len		= max( len( name ) for name in details )
            status		= '\n'.join(
                f"{name:>{name_len}} saved to {values['-TARGET-']}"
                for name in details
            )

    window.close()


def main( argv=None ):
    ap				= argparse.ArgumentParser(
        description = "Create and output SLIP39 encoded Ethereum wallet(s) to a PDF file.",
        formatter_class = argparse.RawDescriptionHelpFormatter,
        epilog = """\

A GUI App for creating SLIP-39 Mnemonic encoded cryptocurrency wallet seeds, either from secure
randomness, or from pre-existing seed entropy, or recovered from prior SLIP-39 or BIP-39 encoded
seed Mnemonics.

This can be useful for converting existing BIP-39 Mnemonic encoded seeds to more secure and
recoverable SLIP-39 Mnemonic encoding.

"""
    )
    ap.add_argument( '-v', '--verbose', action="count",
                     default=0,
                     help="Display logging information." )
    ap.add_argument( '-q', '--quiet', action="count",
                     default=0,
                     help="Reduce logging output." )
    ap.add_argument( '-t', '--threshold',
                     default=None,
                     help="Number of groups required for recovery (default: half of groups, rounded up)" )
    ap.add_argument( '-g', '--group', action='append',
                     help="A group name[[<require>/]<size>] (default: <size> = 1, <require> = half of <size>, rounded up, eg. 'Frens(3/5)' )." )
    ap.add_argument( '-c', '--cryptocurrency', action='append',
                     default=[],
                     help="A crypto name and optional derivation path ('../<range>/<range>' allowed); defaults:" \
                     f" {', '.join( f'{c}:{Account.path_default(c)}' for c in Account.CRYPTOCURRENCIES)}" )
    ap.add_argument( '--passphrase',
                     default=None,
                     help="Encrypt the master secret w/ this passphrase, '-' reads it from stdin (default: None/'')" )
    ap.add_argument( 'names', nargs="*",
                     help="Account names to produce")
    args			= ap.parse_args( argv )

    # Set up logging; also, handle the degenerate case where logging has *already* been set up (and
    # basicConfig is a NO-OP), by (also) setting the logging level
    log_cfg['level']		= log_level( args.verbose - args.quiet )
    logging.basicConfig( **log_cfg )
    if args.verbose:
        logging.getLogger().setLevel( log_cfg['level'] )

    try:
        app(
            names		= args.names,
            threshold		= args.threshold,
            group		= args.group,
            cryptocurrency	= args.cryptocurrency,
            passphrase		= args.passphrase,
        )
    except Exception as exc:
        log.exception( f"Failed running App: {exc}" )
        return 1
    return 0



import argparse
import codecs
import logging

from ..util		import log_cfg, log_level, input_secure, ordinal
from .			import recover, recover_bip39

log				= logging.getLogger( __package__ )


def main( argv=None ):
    ap				= argparse.ArgumentParser(
        description = "Recover and output secret seed from SLIP39 or BIP39 mnemonics",
        formatter_class = argparse.RawDescriptionHelpFormatter,
        epilog = """\
If you obtain a threshold number of SLIP-39 mnemonics, you can recover the original
secret seed, and re-generate one or more Ethereum wallets from it.

Enter the mnemonics when prompted and/or via the command line with -m |--mnemonic "...".

The master secret seed can then be used to generate a new SLIP-39 encoded wallet:

    python3 -m slip39 --secret = "ab04...7f"

BIP-39 wallets can be backed up as SLIP-39 wallets, but only at the cost of 59-word SLIP-39
mnemonics.  This is because the *output* 512-bit BIP-39 seed must be stored in SLIP-39 -- not the
*input* 128-, 160-, 192-, 224-, or 256-bit entropy used to create the original BIP-39 mnemonic
phrase.

""" )

    ap.add_argument( '-v', '--verbose', action="count",
                     default=0,
                     help="Display logging information." )
    ap.add_argument( '-q', '--quiet', action="count",
                     default=0,
                     help="Reduce logging output." )
    ap.add_argument( '-b', '--bip39', action='store_true',
                     default=None,
                     help="Recover 512-bit secret seed from BIP-39 mnemonics" )
    ap.add_argument( '-m', '--mnemonic', action='append',
                     help="Supply another SLIP-39 (or a BIP-39) mnemonic phrase" )
    ap.add_argument( '-p', '--passphrase',
                     default=None,
                     help="Decrypt the master secret w/ this passphrase, '-' reads it from stdin (default: None/'')" )
    args			= ap.parse_args( argv )

    log_cfg['level']		= log_level( args.verbose - args.quiet )

    # Set up logging; also, handle the degenerate case where logging has *already* been set up (and
    # basicConfig is a NO-OP), by (also) setting the logging level
    logging.basicConfig( **log_cfg )
    if args.verbose:
        logging.getLogger().setLevel( log_cfg['level'] )

    # Optional passphrase (utf-8 encoded bytes)
    passphrase			= args.passphrase or ""
    if passphrase == '-':
        passphrase		= input_secure( 'Master seed passphrase: ', secret=True )
    elif passphrase:
        log.warning( "It is recommended to not use '-p|--passphrase <password>'; specify '-' to read from input" )
    passphrase			= passphrase.encode( 'utf-8' )

    # If BIP-39 recovery designated, only a single mnemonic is allowed:
    secret			= None
    algo			= "SLIP-39" if args.bip39 is None else "BIP-39"
    mnemonics			= args.mnemonic or []
    if args.bip39:
        assert 0 <= len( mnemonics ) <= 1, "BIP-39 requires exactly one mnemonic"
        if not mnemonics:
            try:
                phrase		= input_secure( f"Enter {ordinal(len(mnemonics)+1)} {algo} mnemonic: ", secret=False )
            except KeyboardInterrupt:
                return 0
            mnemonics.append( phrase )
        try:
            secret		= recover_bip39( *mnemonics, passphrase )
        except KeyboardInterrupt:
            return 0
        except Exception as exc:
            log.error( f"Could not recover {algo} seed with supplied mnemonic: {exc}" )
    else:
        # Collect more mnemonics 'til we can successfully recover the master secret seed
        while secret is None:
            try:
                secret		= recover( mnemonics, passphrase )
            except KeyboardInterrupt:
                return 0
            except Exception as exc:
                if mnemonics:
                    log.info( f"Could not recover {algo} seed with {len(mnemonics)} supplied mnemonics: {exc}" )
                try:
                    phrase	= input_secure( f"Enter {ordinal(len(mnemonics)+1)} {algo} mnemonic: ", secret=False )
                except KeyboardInterrupt:
                    return 0
                if ':' in phrase:  # Discard any "<name>: <mnemonic>" name prefix.
                    _,phrase	= phrase.split( ':', 1 )
                mnemonics.append( phrase )
    if secret:
        secret			= codecs.encode( secret, 'hex_codec' ).decode( 'ascii' )
        log.info( "Recovered {algo} secret; To re-generate SLIP-39 wallet, send it to: python3 -m slip39 --secret -" )
        print( secret )
    return 0 if secret else 1




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2021, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version. See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

#
# HD Wallet Derivation Paths (Standard BIP-44 / Trezor)
#
#     https://wolovim.medium.com/ethereum-201-hd-wallets-11d0c93c87f7
#
# BIP-44 defines the purpose of each depth level:
#    m / purpose’ / coin_type’ / account’ / change / address_index
#
# Use https://iancoleman.io/bip39/ to confirm the derivations
#

# Default group_threshold / required ratios and groups (with varying styles of definition)
GROUP_REQUIRED_RATIO		= 1/2   # default to 1/2 of group members, rounded up
GROUP_THRESHOLD_RATIO		= 1/2   # default to 1/2 of groups, rounded up
GROUPS				= [
    "First1",
    "Second(1/1)",
    "Fam(4)",
    "Frens3/6"
]

FONTS				= dict(
    sans	= 'helvetica',
    mono	= 'courier',
)

#                                  Y      X       Margin
CREDIT_CARD			= (2+1/4, 3+3/8), 1/16
INDEX_CARD			= (3,     5),     1/8   # noqa: E241
BUSINESS_CARD			= (2,     3+1/2), 1/32  # noqa: E241
HALF_LETTER			= (4+1/2, 8),     1/4   # noqa: E241 (actually, 2/letter, 3/legal)

CARD				= 'index'
CARD_SIZES			= dict(
    index	= INDEX_CARD,
    credit	= CREDIT_CARD,
    business	= BUSINESS_CARD,
    half	= HALF_LETTER,
)

PAGE_MARGIN			= 1/4  # Typical printers cannot print within 1/4" of edge

PAPER				= 'Letter'

BITS				= (128, 256, 512)
BITS_DEFAULT			= 128

MNEM_ROWS_COLS			= {
    20:	( 7, 3),		# 128-bit seed
    33:	(11, 3),		# 256-bit seed
    59:	(12, 5),		# 512-bit seed, eg. from BIP-39 (Unsupported on Trezor hardware wallet)
}


BAUDRATE			= 115200

FILENAME_KEYWORDS		= [ 'name', 'date', 'time', 'crypto', 'path', 'address' ]
FILENAME_FORMAT			= "{name}-{date}+{time}-{crypto}-{address}.pdf"

# Default Crypto accounts (and optional paths) to generate
CRYPTO_PATHS			= [ 'ETH', 'BTC' ]



import pytest

import shamir_mnemonic

from .api		import create
from .recovery		import recover, recover_bip39
from .dependency_test	import substitute, nonrandom_bytes, SEED_XMAS


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_recover():
    details			= create(
        "recovery test", 2, dict( one = (1,1), two = (1,1), fam = (2,4), fren = (3,5) ), SEED_XMAS
    )
    #import json
    #print( json.dumps( details.groups, indent=4 ))
    assert details.groups == {
        "one": (
            1,
            [
                "academic acid acrobat romp chubby client grief judicial pulse domain flip elevator become spirit heat patent hawk remove pickup boring"
            ]
        ),
        "two": (
            1,
            [
                "academic acid beard romp away ancient domain jacket early admit true disaster manual sniff seafood guest stick grumpy blessing unknown"
            ]
        ),
        "fam": (
            2,
            [
                "academic acid ceramic roster density snapshot crush modify born plastic greatest victim merit weapon general cover wits cradle quick emphasis",
                "academic acid ceramic scared brother carve scout stay repeat that fumes tendency junior clay freshman rhyme infant enlarge puny decent",
                "academic acid ceramic shadow class findings zero blessing sidewalk drink jump hormone advocate flip install alpha ugly speak prospect solution",
                "academic acid ceramic sister aluminum obesity blue furl grownup island educate junk traveler listen evidence merit grant python purchase piece"
            ]
        ),
        "fren": (
            3,
            [
                "academic acid decision round academic academic academic academic academic academic academic academic academic academic academic academic academic ranked flame amount",
                "academic acid decision scatter change pleasure dive cricket class impulse lungs hour invasion strike mustang friendly divorce corner penalty fawn",
                "academic acid decision shaft disaster python expand math typical screw rumor research unusual segment install curly debut shadow orange museum",
                "academic acid decision skin browser breathe intimate picture smirk railroad equip spirit nervous capital teaspoon hybrid angel findings hunting similar",
                "academic acid decision snake angel phrase gums response tracks carve secret bucket liquid dictate enemy decrease dance early weapon season"
            ]
        )
    }
    assert recover( details.groups['one'][1] + details.groups['fren'][1][:3] ) == SEED_XMAS

    # Enough correct number of mnemonics must be provided (extras ignored)
    with pytest.raises(shamir_mnemonic.MnemonicError) as excinfo:
        recover( details.groups['one'][1] + details.groups['fren'][1][:2] )
    assert "Wrong number of mnemonics" in str(excinfo.value)

    assert recover( details.groups['one'][1] + details.groups['fren'][1][:4] ) == SEED_XMAS

    # Invalid mnemonic phrases are rejected (one word changed)
    with pytest.raises(shamir_mnemonic.MnemonicError) as excinfo:
        recover( details.groups['one'][1] + details.groups['fren'][1][:2] + [
            "academic acid academic axle crush swing purple violence teacher curly total equation clock mailman display husband tendency smug laundry laundry"
        ])
    assert "Invalid mnemonic checksum" in str(excinfo.value)

    # Duplicate mnemonics rejected/ignored
    with pytest.raises(shamir_mnemonic.MnemonicError) as excinfo:
        recover( details.groups['one'][1] + details.groups['fren'][1][:2] + details.groups['fren'][1][:1] )
    assert "Wrong number of mnemonics" in str(excinfo.value)

    # Mnemonics from another SLIP-39 rejected
    with pytest.raises(shamir_mnemonic.MnemonicError) as excinfo:
        recover( details.groups['one'][1] + details.groups['fren'][1][:2] + [
            "academic acid academic axle crush swing purple violence teacher curly total equation clock mailman display husband tendency smug laundry disaster"
        ])
    assert "Invalid set of mnemonics" in str(excinfo.value)


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_bip39():
    bip39seed			= recover_bip39( 'zoo ' * 11 + 'wrong' )
    details			= create(
        "bip39 recovery test", 2, dict( one = (1,1), two = (1,1), fam = (2,4), fren = (3,5) ),
        master_secret=bip39seed,
    )
    #import json
    #print( json.dumps( details.groups, indent=4 ))
    assert details.groups == {
        "one": (
            1,
            [
                "academic acid acrobat romp academic angel email prospect endorse strategy debris award strike frost actress facility legend safari pistol"
                " mouse hospital identify unwrap talent entrance trust cause ranked should impulse avoid fangs various radar dilemma indicate says rich work"
                " presence jerky glance hesitate huge depend tension loan tolerate news agree geology phrase random simple finger alarm depart inherit grin"
            ]
        ),
        "two": (
            1,
            [
                "academic acid beard romp acne floral cricket answer debris making decorate square withdraw empty decorate object artwork tracks rocky tolerate"
                " syndrome decorate predator sweater ordinary pecan plastic spew facility predator miracle change solution item lizard testify coal excuse lecture"
                " exercise hamster hand crystal rainbow indicate phantom require satisfy flame acrobat detect closet patent therapy overall muscle spill adjust unhappy"
            ]
        ),
        "fam": (
            2,
            [
                "academic acid ceramic roster acquire again tension ugly edge profile custody geology listen hazard smug branch adequate fishing simple adapt fancy"
                " hour method emperor tactics float quiet location satoshi guilt fantasy royal machine dictate squeeze devote oven eclipse writing level sheriff"
                " teacher purchase building veteran spirit woman realize width vanish scholar jewelry desktop stilt random rhyme debut premium theater",
                "academic acid ceramic scared acid space fantasy breathe true recover privacy tactics boring harvest punish swimming leader talent exchange diet"
                " enforce vanish volume organize coastal emperor change intend club scene intimate upgrade dragon burning lily huge market calcium forecast holiday"
                " merit method type ruler equip retailer pancake paces thorn worthy always story promise clock staff floral smart iris repair",
                "academic acid ceramic shadow acne rumor decent elder aspect lizard obesity friendly regular aircraft beyond military campus employer seafood cover"
                " ivory dough galaxy victim diminish average music cause behavior declare brave toxic visual academic include lilac repair morning rapids building"
                " kernel herald careful helpful move hawk flash glimpse seafood listen writing rocky browser change hybrid diet organize system wrote",
                "academic acid ceramic sister academic both legend raspy pecan mixed broken tenant critical again imply finance pacific single echo capital hesitate"
                " piece disease crush slush belong airline smug voice organize dryer standard emission curious charity swing pitch senior behavior vintage chemical"
                " cage editor rebuild costume adult ancestor erode steady makeup depart carpet level sympathy being soldier glimpse airport picture"
            ]
        ),
        "fren": (
            3,
            [
                "academic acid decision round academic academic academic academic academic academic academic academic academic academic academic academic academic"
                " academic academic academic academic academic academic academic academic academic academic academic academic academic academic academic academic"
                " academic academic academic academic academic academic academic academic academic academic academic academic academic academic academic academic"
                " academic academic academic academic academic academic academic aviation endless plastic",
                "academic acid decision scatter acid ugly raspy famous swimming else length gray raspy brother fake aunt auction premium military emphasis perfect"
                " surprise class suitable crunch famous burden military laundry inmate regret elder mixture tenant taught smirk voter process steady artist equip"
                " jury carve acrobat western cylinder gasoline artwork snapshot ancestor object cinema market species platform iris dragon dive medal",
                "academic acid decision shaft acid carbon credit cards rich living humidity peasant source triumph magazine ladle ruin ocean aspect curious round"
                " main evoke deny stadium zero discuss union strike pencil golden silent geology display wrap peanut listen aide learn juice decision plot bike example"
                " obesity ancient square pistol twice sister hour amuse human hobo hospital escape expect wildlife luck",
                "academic acid decision skin academic vanish olympic evoke gesture rumor unfair scroll grasp very steady include smell diploma package guest greatest"
                " firm humidity trial width priest class large photo sniff survive machine usher stick capacity heat improve predator float iris jacket soldier apart"
                " excuse garden cleanup realize permit dough script veteran crazy theater rival secret drink kernel lips pants",
                "academic acid decision snake acid vegan darkness bucket benefit therapy valuable impulse canyon swing distance vampire round losing twin medal treat"
                " amount fiction hush remind faint distance custody device believe campus guest preach mule exhaust regular short phrase column rescue steady float"
                " mixture testify taught fiction usher snake museum detailed agree intend inherit likely typical blimp symbolic prayer course"
            ]
        )
    }
    assert recover( details.groups['one'][1][:] + details.groups['fren'][1][:3] ) == bip39seed



import codecs
import logging

import hdwallet

__all__				= ( "Account", "path_edit" )

log				= logging.getLogger( __package__ )


def path_edit(
    path: str,
    edit: str,
):
    """Replace the current path w/ the new path, either entirely, or if only partially if a continuation
    '.../' followed by some new path segments is provided.

    """
    if edit.startswith( '..' ):
        new_segs	= edit.lstrip( './' ).split( '/' )
        cur_segs	= path.split( '/' )
        log.info( f"Using {edit} to replace last {len(new_segs)} of {path} with {'/'.join(new_segs)}" )
        if len( new_segs ) >= len( cur_segs ):
            raise ValueError( f"Cannot use {edit} to replace last {len(new_segs)} of {path} with {'/'.join(new_segs)}" )
        res_segs	= cur_segs[:len(cur_segs)-len(new_segs)] + new_segs
        return '/'.join( res_segs )
    else:
        return edit


class Account( hdwallet.HDWallet ):

    """Supports producing Legacy addresses for Bitcoin, and Litecoin.  Doge (D...) and Ethereum (0x...)
    addresses use standard BIP44 derivation.


    | Crypto | Semantic | Path             | Address |
    |        |          |                  | <       |
    |--------+----------+------------------+---------|
    | ETH    | Legacy   | m/44'/60'/0'/0/0 | 0x...   |
    | BTC    | Legacy   | m/44'/ 0'/0'/0/0 | 1...    |
    |        | SegWit   | m/44'/ 0'/0'/0/0 | 3...    |
    |        | Bech32   | m/84'/ 0'/0'/0/0 | bc1...  |
    | LTC    | Legacy   | m/44'/ 2'/0'/0/0 | L...    |
    |        | SegWit   | m/44'/ 2'/0'/0/0 | M...    |
    |        | Bech32   | m/84'/ 2'/0'/0/0 | ltc1... |
    | DOGE   | Legacy   | m/44'/ 3'/0'/0/0 | D...    |

    """
    CRYPTOCURRENCIES		= ('ETH', 'BTC', 'LTC', 'DOGE',)  # Currently supported
    CRYPTO_NAMES		= dict(
        ethereum	= 'ETH',
        bitcoin		= 'BTC',
        litecoin	= 'LTC',
        dogecoin	= 'DOGE',
    )

    CRYPTO_FORMAT		= dict(
        ETH		= "legacy",
        BTC		= "bech32",
        LTC		= "bech32",
        DOGE		= "legacy",
    )
    FORMATS		= ("legacy", "segwit", "bech32")

    CRYPTO_FORMAT_PATH		= dict(
        ETH		= dict(
            legacy	= "m/44'/60'/0'/0/0",
        ),
        BTC		= dict(
            legacy	= "m/44'/0'/0'/0/0",
            segwit	= "m/44'/0'/0'/0/0",
            bech32	= "m/84'/0'/0'/0/0",
        ),
        LTC		= dict(
            legacy	= "m/44'/2'/0'/0/0",
            segwit	= "m/44'/2'/0'/0/0",
            bech32	= "m/84'/2'/0'/0/0",
        ),
        DOGE		= dict(
            legacy	="m/44'/3'/0'/0/0",
        )
    )

    @classmethod
    def path_default( cls, crypto, format=None ):
        """Return the default derivation path for the given crypto, based on its currently selected default
        address format.

        """
        crypto			= cls.supported( crypto )
        format			= format.lower() if format else cls.address_format( crypto )
        if format not in cls.CRYPTO_FORMAT_PATH[crypto]:
            raise ValueError( f"{format} not supported for {crypto}; specify one of {', '.join( cls.CRYPTO_FORMAT_PATH[crypto].keys() )}" )
        return cls.CRYPTO_FORMAT_PATH[crypto][format]

    @classmethod
    def address_format( cls, crypto, format=None ):
        """Get or set the desired default address format for the specified supported crypto.  Future
        instances of Address created for the crypto will use the specified address format and its
        default derivation path.

        """
        crypto			= cls.supported( crypto )
        if format is None:
            return cls.CRYPTO_FORMAT[crypto]

        format			= format.lower() if format else None
        if format not in cls.FORMATS:
            raise ValueError( f"{crypto} address format {format!r} not recognized; specify one of {', '.join( cls.FORMATS )}" )
        cls.CRYPTO_FORMAT[crypto]	= format

    @classmethod
    def supported( cls, crypto ):
        """Validates that the specified cryptocurrency is supported and returns the normalized short name
        for it, or raises an a ValueError.  Eg. "Ethereum" --> "ETH"

        """
        validated			= cls.CRYPTO_NAMES.get(
            crypto.lower(),
            crypto.upper() if crypto.upper() in cls.CRYPTOCURRENCIES else None
        )
        if validated:
            return validated
        raise ValueError( f"{crypto} not presently supported; specify {', '.join( cls.CRYPTOCURRENCIES )}" )

    def __init__( self, crypto, format=None ):
        crypto			= Account.supported( crypto )
        self.format		= format.lower() if format else Account.address_format( crypto )
        if self.format in ("legacy", "segwit",):
            self.hdwallet	= hdwallet.BIP44HDWallet( symbol=crypto )
        elif self.format in ("bech32",):
            self.hdwallet	= hdwallet.BIP84HDWallet( symbol=crypto )
        else:
            raise ValueError( f"{crypto} does not support address format {self.format}" )

    def from_seed( self, seed: str, path: str = None ) -> "Account":
        """Derive the Account from the supplied seed and (optionally) path; uses the default derivation path
        for the Account address format, if None provided.

        """
        if type( seed ) is bytes:
            seed		= codecs.encode( seed, 'hex_codec' ).decode( 'ascii' )
        self.hdwallet.from_seed( seed )
        self.from_path( path )
        return self

    def from_path( self, path: str = None ) -> "Account":
        """Change the Account to derive from the provided path.

        If a partial path is provided (eg "...1'/0/3"), then use it to replace the given segments in
        current account path, leaving the remainder alone.

        """
        if path:
            path		= path_edit( self.path, path )
        else:
            path		= Account.path_default( self.crypto, self.format )
        self.hdwallet.clean_derivation()
        self.hdwallet.from_path( path )
        return self

    @property
    def address( self ):
        if self.format == "legacy":
            return self.hdwallet.p2pkh_address()
        elif self.format == "segwit":
            return self.hdwallet.p2sh_address()
        elif self.format == "bech32":
            return self.hdwallet.p2wpkh_address()
        raise ValueError( f"Unknown addresses semantic: {self.format}" )

    @property
    def crypto( self ):
        return self.hdwallet._cryptocurrency.SYMBOL

    @property
    def path( self ):
        return self.hdwallet.path()

    @property
    def key( self ):
        return self.hdwallet.private_key()



import codecs
import json
import math

import pytest

import shamir_mnemonic

from .api		import create, group_parser
from .recovery		import recover
from .defaults		import GROUPS, GROUP_THRESHOLD_RATIO

from .dependency_test	import substitute, nonrandom_bytes

SEED_FF_HEX			= 'ff' * 16
SEED_FF				= codecs.decode( SEED_FF_HEX, 'hex_codec' )


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_passphrase():
    """Confirm that the default ff x 16 seed w/ no passphrase results in the expected default SLIP39
    Mnemonics, and the expected ETH 0x824b... and BTC bc1q9y... wallets.  These are confirmed as
    being consistent w/ the Trezor.

    Also confirm that we implement SLIP39 passphrases correctly, compared to the Satoshi Labs:

        $ shamir recover -p
        Enter a recovery share: academic acid acrobat romp dining employer endless speak main stick database tenant alarm wealthy plan amuse credit crush arcade cylinder

        Completed 1 of 2 groups needed:
        ✓ 1 of 1 shares needed from group academic acid acrobat
        ✗ 0 shares from group academic acid beard
        ✗ 0 shares from group academic acid ceramic
        ✗ 0 shares from group academic acid decision
        Enter a recovery share: academic acid beard romp dismiss mayor juice glance twice response eclipse inmate muscle climate born increase reject typical slow evil

        Completed 2 of 2 groups needed:
        ✓ 1 of 1 shares needed from group academic acid acrobat
        ✓ 1 of 1 shares needed from group academic acid beard
        ✗ 0 shares from group academic acid ceramic
        ✗ 0 shares from group academic acid decision
        Enter passphrase: password
        Repeat for confirmation: password
        SUCCESS!
        Your master secret is: ffffffffffffffffffffffffffffffff

    """
    groups			= dict(
        group_parser( g )
        for g in GROUPS
    )
    group_threshold		= math.ceil( len( groups ) * GROUP_THRESHOLD_RATIO )

    details_nonpass		= create(
        name		= "SLIP39 Wallet: FF",
        group_threshold	= group_threshold,
        groups		= groups,
        master_secret	= SEED_FF,
    )
    print( json.dumps( details_nonpass.groups, indent=4 ))
    assert details_nonpass.groups == {
        "First": (
            1,
            [
                "academic acid acrobat romp change injury painting safari drug browser trash fridge busy finger standard angry similar overall prune ladybug"
            ]
        ),
        "Second": (
            1,
            [
                "academic acid beard romp believe impulse species holiday demand building earth warn lunar olympic clothes piece campus alpha short endless"
            ]
        ),
        "Fam": (
            2,
            [
                "academic acid ceramic roster desire unwrap depend silent mountain agency fused primary clinic alpha database liberty silver advance replace medical",
                "academic acid ceramic scared column screw hawk dining invasion bumpy identify anxiety august sunlight intimate satoshi hobo traveler carbon class",
                "academic acid ceramic shadow believe revenue type class station domestic already fact desktop penalty omit actress rumor beaver forecast group",
                "academic acid ceramic sister actress mortgage random talent device clogs craft volume cargo item scramble easy grumpy wildlife wrist simple"
            ]
        ),
        "Frens": (
            3,
            [
                "academic acid decision round academic academic academic academic academic academic academic academic academic academic academic academic academic ranked flame amount",
                "academic acid decision scatter biology trial escape element unfair cage wavy afraid provide blind pitch ultimate hybrid gravity formal voting",
                "academic acid decision shaft crunch glance exclude stilt grill numb smug stick obtain raisin force theater duke taught license scramble",
                "academic acid decision skin disaster mama alive nylon mansion listen cowboy suitable crisis pancake velvet aviation exhaust decent medal dominant",
                "academic acid decision snake aunt frozen flip crystal crystal observe equip maximum maiden dragon wine crazy nervous crystal profile fiction",
                "academic acid decision spider bulge receiver behavior humidity ruler luck public math distance cylinder fantasy seafood training withdraw payment playoff"
            ]
        )
    }

    [(eth,btc)] 		= details_nonpass.accounts
    assert eth.path == "m/44'/60'/0'/0/0"
    assert eth.address == '0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1'
    assert btc.path == "m/84'/0'/0'/0/0"
    assert btc.address == 'bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl'

    # Ensure we can recover it w/ no passphrase
    assert recover(
        details_nonpass.groups['Fam'][1][:2] + details_nonpass.groups['Frens'][1][:-3],
    ) == SEED_FF

    # Now, ensure that we see a different set of Mnemonics w/ a SLIP39 passphrase.

    # NOTE: this is NOT what the Trezor uses when you enter a "hidden" wallet passphrase!  There is
    # no support for SLIP39 passphrase decryption on the Trezor.  However, confirm that it works as
    # specified by the shamir_mnemonic module.

    # First, ensure we do *not* recover the origin seed when we use an incorrect passphrase with the
    # Mnemonics generated with *not* passphrase.

    badpass			= "password".encode( 'UTF-8' )

    assert recover(
        details_nonpass.groups['Fam'][1][:2] + details_nonpass.groups['Frens'][1][:-3],
        passphrase	= badpass,
    ) != SEED_FF

    details_badpass		= create(
        name		= "SLIP39 Wallet: FF",
        group_threshold	= group_threshold,
        groups		= groups,
        master_secret	= SEED_FF,
        passphrase	= badpass,
    )
    print( json.dumps( details_badpass.groups, indent=4 ))
    assert details_badpass.groups == {
        "First": (
            1,
            [
                "academic acid acrobat romp dining employer endless speak main stick database tenant alarm wealthy plan amuse credit crush arcade cylinder"
            ]
        ),
        "Second": (
            1,
            [
                "academic acid beard romp dismiss mayor juice glance twice response eclipse inmate muscle climate born increase reject typical slow evil"
            ]
        ),
        "Fam": (
            2,
            [
                "academic acid ceramic roster answer charity ceiling mason kernel jerky woman laundry duckling dining dress grumpy lily fortune cleanup black",
                "academic acid ceramic scared dwarf quarter adequate umbrella inform hamster grief album broken humidity lecture domain estimate losing revenue mouse",
                "academic acid ceramic shadow cluster cards artwork wrap grief forbid skin ordinary vocal lunar engage sister sugar pink maximum cover",
                "academic acid ceramic sister benefit marathon database ordinary hawk equation fawn thorn soldier taste welcome mild blimp insect born pistol"
            ]
        ),
        "Frens": (
            3,
            [
                "academic acid decision round academic academic academic academic academic academic academic academic academic academic academic academic academic ranked flame amount",
                "academic acid decision scatter become flavor crystal genuine hour infant voice unfair recall living afraid company froth distance bundle soldier",
                "academic acid decision shaft display headset level prevent verdict genius preach glad makeup element twin scene hunting smoking rival mortgage",
                "academic acid decision skin course dress prisoner skunk ordinary blind freshman member drink scroll traffic thank deliver frequent velvet evoke",
                "academic acid decision snake burden unknown guard peaceful artwork snapshot teaspoon literary move spray saver voice august maximum hazard negative",
                "academic acid decision spider ancestor regular evaluate salon glasses penalty blue guitar check extra roster snapshot fantasy adjust coastal eclipse"
            ]
        )
    }

    # Now, without the correct password, we can't recover the original SEED_FF.  We *do* recover "a"
    # seed: there is no verification/validation that the supplied passphrase was "correct".  You
    # simply are expected to know that the derived wallet(s) are the ones you expect.  This is by
    # design -- you can supply an attacker with a "duress" seed passphrase leading to wallet(s) with
    # a small sacrificial amount of funds.
    assert recover(
        details_badpass.groups['Fam'][1][:2] + details_badpass.groups['Frens'][1][:3],
    ) != SEED_FF
    assert recover(
        details_badpass.groups['Fam'][1][:2] + details_badpass.groups['Frens'][1][:3],
        passphrase	= badpass,
    ) == SEED_FF

    # Mixing SLIP39 recovery groups should fail to recover, both without and with the password,
    # since SLIP39 confirms the digest of the recovered "encrypted" seed, before decryption.
    with pytest.raises( shamir_mnemonic.utils.MnemonicError ):
        assert recover(
            details_nonpass.groups['Fam'][1][:2] + details_badpass.groups['Frens'][1][:3],
        ) != SEED_FF
    with pytest.raises( shamir_mnemonic.utils.MnemonicError ):
        assert recover(
            details_nonpass.groups['Fam'][1][:2] + details_badpass.groups['Frens'][1][:3],
            passphrase	= badpass,
        ) != SEED_FF

    # And finally, confirm that the SLIP39 Mnemonics protected by the correct passphrase also yield
    # the expected ETH and BTC wallets.  This is obviously the expected behaviour, because the
    # SLIP39 passphrase simply encrypts/decrypts the original seed.
    [(eth,btc)] 		= details_badpass.accounts
    assert eth.path == "m/44'/60'/0'/0/0"
    assert eth.address == '0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1'
    assert btc.path == "m/84'/0'/0'/0/0"
    assert btc.address == 'bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl'



import os
import logging
import pty
import threading

import pytest

try:
    from serial		import Serial
except ImportError:
    Serial			= None

from .api		import RANDOM_BYTES, accountgroups
from .generator		import chacha20poly1305, accountgroups_output, accountgroups_input


def listener(port):
    # continuously listen to commands on the master device
    while True:
        res = b""
        while not res.endswith(b"\r\n"):
            # keep reading one byte at a time until we have a full line
            res += os.read(port, 1)
        print("command: %s" % res)

        # write back the response
        if res == b'QPGS\r\n':
            os.write(port, b"correct result\r\n")
        else:
            os.write(port, b"I don't understand\r\n")


@pytest.mark.skipif( not Serial,
                     reason="Serial testing needs pyserial" )
def test_serial():
    """Start the testing"""
    master,slave = pty.openpty()		# open the pseudoterminal
    s_name = os.ttyname(slave)			# translate the slave fd to a filename

    logging.info( f"Pty name: {s_name}" )

    # create a separate thread that listens on the master device for commands
    thread = threading.Thread(target=listener, args=[master])
    thread.daemon = True
    thread.start()

    # open a pySerial connection to the slave.  Receiver can signal sender to stop/start
    ser = Serial(s_name, 2400, timeout=1)

    ser.write(b'test2\r\n')			# write the first command
    res = b""
    while not res.endswith(b'\r\n'):
        # read the response
        res += ser.read()
    print("result: %s" % res)
    assert res.startswith( b"I don't understand" )

    ser.write(b'QPGS\r\n')			# write a second command
    res = b""
    while not res.endswith(b'\r\n'):
        # read the response
        res += ser.read()
    print("result: %s" % res)
    assert res.startswith( b"correct result" )


def generator( password, cryptopaths, fd ):
    """Generate a sequence of Accounts to the given file descriptor."""
    fdout		= os.fdopen( fd, "w" )
    nonce		= RANDOM_BYTES( 12 )
    cipher		= chacha20poly1305( password=password )
    for index,group in enumerate( accountgroups(
        master_secret	= b'\xff' * 16,
        cryptopaths	= cryptopaths,
    )):
        logging.info( f"Sending: {group}" )
        accountgroups_output(
            group	= group,
            index	= index,
            cipher	= cipher,
            nonce	= nonce,
            corrupt	= .01,
            file	= fdout,
        )


@pytest.mark.skipif( not Serial,
                     reason="Serial testing needs pyserial" )
def test_groups_pty():
    password			= "password"
    master,slave		= pty.openpty()
    slave_name			= os.ttyname( slave )

    cryptopaths			= [
        ("ETH", "m/44'/60'/0'/0/-10"),
        ("ETH", "m/44'/ 0'/0'/0/-10"),
    ]

    gen				= threading.Thread(
        target	= generator,
        args	= [password, cryptopaths, master] )
    gen.daemon			= True
    gen.start()

    class SerialEOF( Serial ):
        """Convert any SerialException to an EOFError, for compatibility with PTY.  In real serial
        ports, we'll handle detection of counterparty readiness with DTR/DSR, and flow control with
        RTS/CTS.

        """
        def read( self, size=1 ):
            while True:
                try:
                    return super( SerialEOF, self ).read( size=size )
                except Exception as exc:  # SerialError as exc:
                    # if "readiness" in str(exc):
                    #     time.sleep( .1 )
                    #     continue
                    raise EOFError( str( exc ))

    ser				= SerialEOF( slave_name, timeout=1)
    for group in accountgroups_input(
        cipher		= chacha20poly1305( password=password ),
        encoding	= 'UTF-8',
        file		= ser
    ):
        logging.info( f"Receive: {group}" )


# def transmitter():


#     yield from addresses(
#         master_secret			= b"\xFF" * 16,
#         paths


# def test_serial_xonxoff():
#     """Ensure that flow-control works.  The idealy secure solution is to use hardware RTS/CTS flow
#     control, with no tx circuit.  However, even a regular serial link with software XON/XOFF flow
#     control is much more secure than a network-connected peer."""



import itertools
import logging

from typing		import List

from shamir_mnemonic	import combine_mnemonics
from mnemonic		import Mnemonic

from ..util		import ordinal
log				= logging.getLogger( __package__ )


def recover(
    mnemonics: List[str],
    passphrase: bytes		= b"",
) -> bytes:
    """Recover a master secret from the supplied SLIP-39 mnemonics.  We cannot know what subset of
    these mnemonics is required and/or valid, so we need to iterate over all subset combinations on
    failure.

    We'll try to find the smallest subset that satisfies the SLIP39 recovery.

    """
    secret			= None
    try:
        combo			= range( len( mnemonics ))
        secret			= combine_mnemonics( mnemonics, passphrase=passphrase )
    except Exception as exc:
        # Try a subset of the supplied mnemonics, to silently reject any invalid mnemonic phrases supplied
        for length in range( len( mnemonics )):
            for combo in itertools.combinations( range( len( mnemonics )), length ):
                trial		= list( mnemonics[i] for i in combo )
                try:
                    secret	= combine_mnemonics( trial, passphrase=passphrase )
                    break
                except Exception:
                    pass
            if secret:
                break
        if not secret:
            # No recovery; raise the Exception produced by original attempt w/ all mnemonics
            raise exc
    log.warning( f"Recovered {len(secret)*8}-bit SLIP-39 secret with {len(combo)}"
                 f" ({'all' if len(combo) == len(mnemonics) else ', '.join( ordinal(i+1) for i in combo)}) "
                 f"of {len(mnemonics)} supplied mnemonics" )
    return secret


def recover_bip39(
    mnemonic: str,
    passphrase: bytes		= b"",
) -> bytes:
    """Recover a 512-bit seed from a single BIP-39 mnemonic phrase, detecting the language."""
    language			= Mnemonic.detect_language( mnemonic )
    m				= Mnemonic( language )
    assert m.check( mnemonic ), \
        f"Invalid BIP-39 mnemonic: {mnemonic}"
    secret			= Mnemonic.to_seed( mnemonic, passphrase )
    log.warning( f"Recovered {len(secret)*8}-bit BIP-39 secret from {language} mnemonic" )
    return secret



import json

import shamir_mnemonic

from .api		import account, create, addresses, addressgroups, accountgroups
from .recovery		import recover

from .dependency_test	import substitute, nonrandom_bytes, SEED_XMAS


def test_account():
    acct			= account( SEED_XMAS )
    assert acct.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert acct.path == "m/44'/60'/0'/0/0"

    acct			= account( SEED_XMAS, path="m/44'/60'/0'/0/1" )
    assert acct.address == '0x3b774e485fC818F0f377FBA657dfbF92B46f8504'
    assert acct.path == "m/44'/60'/0'/0/1"

    acct			= account( SEED_XMAS, crypto='Bitcoin' )
    assert acct.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'
    assert acct.path == "m/84'/0'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Bitcoin', format='Legacy' )
    assert acct.address == '19FQ983heQEBXmopVNyJKf93XG7pN7sNFa'
    assert acct.path == "m/44'/0'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Bitcoin', format='SegWit' )
    assert acct.address == '3HxUpD7E8Y31vDDgDq1VFdNXWViAgBjYJe'
    assert acct.path == "m/44'/0'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Litecoin' )
    assert acct.address == 'ltc1qfjepkelqd3jx4e73s7p79lls6kqvvmak5pxy97'
    assert acct.path == "m/84'/2'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Litecoin', format='Legacy' )
    assert acct.address == 'LeyK1dbc5qKdKC9TvkygMTeoHixR3z1XG3'
    assert acct.path == "m/44'/2'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Litecoin', format='SegWit' )
    assert acct.address == 'MPULjvY9dNjpNkgbwhfJtD7N6Lbfam1XsP'
    assert acct.path == "m/44'/2'/0'/0/0"

    acct			= account( SEED_XMAS, crypto='Dogecoin' )
    assert acct.address == 'DQCnF49GwQ5auL3u5c2uow62XS57RCA2r1'
    assert acct.path == "m/44'/3'/0'/0/0"


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_create():
    details			= create(
        "SLIP39 Wallet: Test",
        1,
        dict( fren = (3,5) ),
        SEED_XMAS
    )

    assert details.groups == {
        "fren": ( 3, [
            "academic acid academic acne academic academic academic academic academic academic academic academic academic academic academic academic academic carpet making building",
            "academic acid academic agree depart dance galaxy acrobat mayor disaster quick justice ordinary agency plunge should pupal emphasis security obtain",
            "academic acid academic amazing crush royal faint spit briefing craft floral negative work depend prune adapt merit romp home elevator",
            "academic acid academic arcade cargo unfold aunt spider muscle bedroom triumph theory gather dilemma building similar chemical object cinema salon",
            "academic acid academic axle crush swing purple violence teacher curly total equation clock mailman display husband tendency smug laundry disaster"
        ] ),
    }

    assert len(details.accounts) == 1
    [(eth,btc)] = details.accounts  # The default accounts created are ETH, BTC
    assert eth.address == '0x336cBeAB83aCCdb2541e43D514B62DC6C53675f4'
    assert btc.address == 'bc1qz6kp20ukkyx8c5t4nwac6g8hsdc5tdkxhektrt'

    assert recover( details.groups['fren'][1][:3] ) == SEED_XMAS


def test_addresses():
    master_secret		= b'\xFF' * 16
    addrs			= list( addresses(
        master_secret	= master_secret,
        crypto		= 'ETH',
        paths		= ".../-9",
    ))
    # print( json.dumps( addrs, indent=4, default=str ))
    assert addrs == [
        (
            "ETH",
            "m/44'/60'/0'/0/0",
            "0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/1",
            "0x8D342083549C635C0494d3c77567860ee7456963"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/2",
            "0x52787E24965E1aBd691df77827A3CfA90f0166AA"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/3",
            "0xc2442382Ae70c77d6B6840EC6637dB2422E1D44e"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/4",
            "0x42a910D380dE132B5227e3277Cc70C3C76a884aC"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/5",
            "0x1A3db5E0422c78F43a35686f0307Da8f22344dE0"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/6",
            "0x19031c515C5d91DB7988D89AAA6F71a5825f5245"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/7",
            "0xaE693156ac600f5B0D58e5090ecf0A578c5Cc0a8"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/8",
            "0x4347541fa648BCE62543a8AbC2901E08017f6A6a"
        ),
        (
            "ETH",
            "m/44'/60'/0'/0/9",
            "0xC11235559Dd4c5224a19396C3A14526E92ebba35"
        )
    ]
    addrs			= list( addresses(
        master_secret	= master_secret,
        crypto		= 'BTC',
        paths		= '.../0/-2',
        format		= 'Bech32',
    ))
    print( json.dumps( addrs, indent=4, default=str ))
    assert addrs == [
        (
            "BTC",
            "m/84'/0'/0'/0/0",
            "bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl"
        ),
        (
            "BTC",
            "m/84'/0'/0'/0/1",
            "bc1qnec684yvuhfrmy3q856gydllsc54p2tx9w955c"
        ),
        (
            "BTC",
            "m/84'/0'/0'/0/2",
            "bc1q2snj0zcg23dvjpw7m9lxtu0ap0hfl5tlddq07j"
        ),
    ]


def test_addressgroups():
    master_secret		= b'\xFF' * 16
    addrgrps			= list( enumerate( addressgroups(
        master_secret	= master_secret,
        cryptopaths	= [
            ('ETH', ".../-3"),
            ('BTC', ".../-3"),
            ('LTC', ".../-3"),
            ('Doge', ".../-3"),
        ],
    )))
    # print( addrgrps )
    assert addrgrps == [
        (0, (("ETH", "m/44'/60'/0'/0/0", "0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1"),
             ("BTC", "m/84'/0'/0'/0/0", "bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl"),
             ('LTC', "m/84'/2'/0'/0/0", 'ltc1qe5m2mst9kjcqtfpapaanaty40qe8xtusmq4ake'),
             ('DOGE', "m/44'/3'/0'/0/0", 'DN8PNN3dipSJpLmyxtGe4EJH38EhqF8Sfy'))),
        (1, (("ETH", "m/44'/60'/0'/0/1", "0x8D342083549C635C0494d3c77567860ee7456963"),
             ("BTC", "m/84'/0'/0'/0/1", "bc1qnec684yvuhfrmy3q856gydllsc54p2tx9w955c"),
             ('LTC', "m/84'/2'/0'/0/1", 'ltc1qm0hwvvk28wlyfu3sed66e9yyvmwm35xtfexva3'),
             ('DOGE', "m/44'/3'/0'/0/1",'DJYE9WWaCA1CbV9x23qkcgNX7Yr9YjCebA'))),
        (2, (("ETH", "m/44'/60'/0'/0/2", "0x52787E24965E1aBd691df77827A3CfA90f0166AA"),
             ("BTC", "m/84'/0'/0'/0/2", "bc1q2snj0zcg23dvjpw7m9lxtu0ap0hfl5tlddq07j"),
             ('LTC', "m/84'/2'/0'/0/2", 'ltc1qx3r3efsmupn34gmwu25fu39tn4h79cjfwvlpfu'),
             ('DOGE', "m/44'/3'/0'/0/2",'DQfJcJzLFW9LJPJXNkLeq1WqPfLsRq47Jj'))),
        (3, (("ETH", "m/44'/60'/0'/0/3", "0xc2442382Ae70c77d6B6840EC6637dB2422E1D44e"),
             ("BTC", "m/84'/0'/0'/0/3", "bc1qxwekjd46aa5n0s3dtsynvtsjwsne7c5f5w5dsd"),
             ('LTC', "m/84'/2'/0'/0/3", 'ltc1qnqzyear8kct0yjzupe2pxtq0mwee5kl642mj78'),
             ('DOGE', "m/44'/3'/0'/0/3", 'DLVPiM5763cyNJfoa13cv4kV3b87FgVMCS'))),
    ]


def test_accountgroups():
    master_secret		= b'\xFF' * 16
    acctgrps			= list( accountgroups(
        master_secret	= master_secret,
        cryptopaths	= [
            ('ETH', ".../-3"),
            ('BTC', ".../-3"),
        ],
    ))
    # print( json.dumps( acctgrps, default=repr ))
    assert len(acctgrps) == 4



import itertools
import logging
import math
import re
import secrets

from collections	import namedtuple
from typing		import Dict, List, Sequence, Tuple, Union, Callable

from shamir_mnemonic	import generate_mnemonics

from .types		import Account
from .defaults		import BITS_DEFAULT, BITS, MNEM_ROWS_COLS, GROUP_REQUIRED_RATIO
from .util		import ordinal


RANDOM_BYTES			= secrets.token_bytes

log				= logging.getLogger( __package__ )


def path_parser(
    paths: str,
    allow_unbounded: bool	= True,
) -> Tuple[str, Dict[str, Callable[[], int]]]:
    """Create a format and a dictionary of iterators to feed into it.

    Supports paths with an arbitrary prefix, eg. 'm/' or '.../'
    """
    path_segs			= paths.split( '/' )
    unbounded			= False
    ranges			= {}

    for i,s in list( enumerate( path_segs )):
        if '-' not in s:
            continue
        c			= chr(ord('a')+i)
        tic			= s.endswith( "'" )
        if tic:
            s			= s[:-1]
        b,e			= s.split( '-' )
        b			= int( b or 0 )
        if e:
            e			= int( e )
            ranges[c]		= lambda b=b,e=e: range( b, e+1 )
        else:
            assert allow_unbounded and not ( unbounded or ranges ), \
                f"{'Only first' if allow_unbounded else 'No'} range allowed to be unbounded;" \
                f" this is the {ordinal(len(ranges)+1)} range in {paths}"
            unbounded		= True
            ranges[c]		= lambda b=b: itertools.count( b )
        path_segs[i]		= f"{{{c}}}" + ( "'" if tic else "" )

    path_fmt			= '/'.join( path_segs )
    return path_fmt, ranges


def path_sequence(
    path_fmt: str,
    ranges: Dict[str, Callable[[], int]],
):
    """Yield a sequence of paths, modulating the format specifiers of the
    path_fmt according to their value sources in ranges.

    For example, a

        path_fmt = "m/44'/60'/0'/0/{f}", with a
        ranges   = dict( f=lambda b=0, e=2: range( b, e+1 ) )

    would yield the paths:

        "m/44'/60'/0'/0/0"
        "m/44'/60'/0'/0/1"
        "m/44'/60'/0'/0/2"
    """
    # Stert all the iterators
    viters			= {
        k: iter( l() )
        for k,l in ranges.items()
    }
    values			= {		# Initial round of values; must provide at least one
        k: next( viters[k] )
        for k in viters
    }
    assert all( v is not None for v in values.values() ), \
        "Iterators for path segment values must yield at least an initial value"

    while not any( v is None for v in values.values() ):
        yield path_fmt.format( **values )
        if not ranges:
            break				# No variable records at all; just one
        # Get the next value.  Working from the lowest iterator up, cycle value(s)
        for i,k in enumerate( sorted( viters.keys(), reverse=True )):
            values[k]		= next( viters[k], None )
            if values[k] is not None:
                break
            # OK, this iterable has ended.  Restart it, and cycle to the next one up, iff
            # there are remaining ranges
            if i+1 < len( ranges ):
                viters[k]	= iter( ranges[k]() )
                values[k]	= next( viters[k], None )


def random_secret(
    seed_length			= BITS_DEFAULT // 8
) -> bytes:
    """Generates a new random secret.

    NOTE: There is a slightly less than 1 / 2^128 chance that any given random secret will lead to
    an invalid BTC wallet private key!  This is because the 256-bit seed for bitcoin must be less than
    the secp256k1 field size:

        0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141

    We cannot generate secrets that are guaranteed to be valid for every derived HD Wallet BTC
    address.  The secret is derived through a complex hashing procedure for each wallet at each
    path.  The probability of this occurring is so vanishingly small that we will instead simply opt
    to *not* generate an public wallet address, when asked.

    Just to give you an understandable reference, the first 128 bits of an illegal secret unusable
    for generating a Bitcoin wallet must have 128 '1' bits at its start.  If wallets were generated
    at a billion per second, by every person alive on earth, it should take about 1.5 trillion years
    to arrive at the first invalid seed.

    2**128 / 1e9 / 7e9 / (60*60*24*365) == 1,541,469,010,115.145

    """
    return RANDOM_BYTES( seed_length )


Details = namedtuple( 'Details', ('name', 'group_threshold', 'groups', 'accounts') )


def enumerate_mnemonic( mnemonic ):
    if isinstance( mnemonic, str ):
        mnemonic		= mnemonic.split( ' ' )
    return dict(
        (i, f"{i+1:>2d} {w}")
        for i,w in enumerate( mnemonic )
    )


def organize_mnemonic( mnemonic, rows=None, cols=None, label="" ):
    """Given a SLIP-39 "word word ... word" or ["word", "word", ..., "word"] mnemonic, emit rows
    organized in the desired rows and cols (with defaults, if not provided).  We return the fully
    formatted line, plus the list of individual words in that line.

    """
    num_words			= enumerate_mnemonic( mnemonic )
    if not rows or not cols:
        rows,cols		= MNEM_ROWS_COLS.get( len(num_words), (7, 3))
    for r in range( rows ):
        line			= label if r == 0 else ' ' * len( label )
        words			= []
        for c in range( cols ):
            word		= num_words.get( c * rows + r )
            if word:
                words.append( word )
                line	       += f"{word:<13}"
        yield line,words


def group_parser( group_spec ):
    """Parse a SLIP-39 group specification.

        Frens6, Frens 6, Frens(6)	- A 3/6 group (default is 1/2 of group size, rounded up)
        Frens2/6, Frens(2/6)		- A 2/6 group

    """
    match			= group_parser.RE.match( group_spec )
    if not match:
        raise ValueError( f"Invalid group specification: {group_spec!r}" )
    name			= match.group( 'name' )
    size			= match.group( 'size' )
    require			= match.group( 'require' )
    if not size:
        size			= 1
    if not require:
        # eg. default 2/4, 3/5
        require			= math.ceil( int( size ) * GROUP_REQUIRED_RATIO )
    return name,(int(require),int(size))
group_parser.RE			= re.compile( # noqa E305
    r"""^
        \s*
        (?P<name> [^\d\(/]+ )
        \s*\(?\s*
        (:? (?P<require> \d* ) \s* / )?
        \s*
        (?P<size> \d* )
        \s*\)?\s*
        $""", re.VERBOSE )


def create(
    name: str,
    group_threshold: int,
    groups: Dict[str,Tuple[int, int]],
    master_secret: bytes	= None,		# Default: 128-bit seeds
    passphrase: bytes		= b"",
    iteration_exponent: int	= 1,
    cryptopaths: Sequence[Tuple[str,str]] = None,  # default: ETH, BTC at default paths
    strength: int		= 128,
) -> Tuple[str,int,Dict[str,Tuple[int,List[str]]], Sequence[Sequence[Account]]]:
    """Creates a SLIP-39 encoding and 1 or more Ethereum accounts.  Returns the details, in a form
    compatible with the output API.

    """
    if master_secret is None:
        assert strength in BITS, f"Invalid {strength}-bit secret length specified"
        master_secret		= random_secret( strength // 8 )
    g_names,g_dims		= list( zip( *groups.items() ))
    mnems			= mnemonics(
        group_threshold	= group_threshold,
        groups		= g_dims,
        master_secret	= master_secret,
        passphrase	= passphrase,
        iteration_exponent= iteration_exponent
    )
    # Derive the desired account(s) at the specified derivation paths, or the default
    accts			= list( accountgroups(
        master_secret	= master_secret,
        cryptopaths	= cryptopaths or [('ETH',None), ('BTC',None)],
        allow_unbounded	= False,
    ))

    groups			= {
        g_name: (g_of, g_mnems)
        for (g_name,(g_of, _),g_mnems) in zip( g_names, g_dims, mnems )
    }
    if log.isEnabledFor( logging.INFO ):
        group_reqs			= list(
            f"{g_nam}({g_of}/{len(g_mns)})" if g_of != len(g_mns) else f"{g_nam}({g_of})"
            for g_nam,(g_of,g_mns) in groups.items() )
        requires		= f"Recover w/ {group_threshold} of {len(groups)} groups {', '.join(group_reqs)}"
        for g_n,(g_name,(g_of,g_mnems)) in enumerate( groups.items() ):
            log.info( f"{g_name}({g_of}/{len(g_mnems)}): {requires}" )
            for mn_n,mnem in enumerate( g_mnems ):
                for line,_ in organize_mnemonic( mnem, label=f"{ordinal(mn_n+1)} " ):
                    log.info( f"{line}" )

    return Details(name, group_threshold, groups, accts)


def mnemonics(
    group_threshold: int,
    groups: Sequence[Tuple[int, int]],
    master_secret: Union[str,bytes] = None,
    passphrase: bytes		= b"",
    iteration_exponent: int	= 1,
    strength: int		= 128,
) -> List[List[str]]:
    """Generate SLIP39 mnemonics for the supplied group_threshold of the given groups.  Will generate a
     random master_secret, if necessary.

    """
    if master_secret is None:
        assert strength in BITS, f"Invalid {strength}-bit secret length specified"
        master_secret		= random_secret( strength // 8 )
    if len( master_secret ) * 8 not in BITS:
        raise ValueError(
            f"Only 128-, 256- and 512bit seeds supported; {len(master_secret)*8}-bit master_secret supplied" )
    return generate_mnemonics(
        group_threshold	= group_threshold,
        groups		= groups,
        master_secret	= master_secret,
        passphrase	= passphrase,
        iteration_exponent = iteration_exponent )


def account(
    master_secret: Union[str,bytes],
    crypto: str			= None,  # default 'ETH'
    path: str			= None,  # default to the crypto's path_default
    format: str			= None,  # eg. 'bech32', or use the default address format for the crypto
):
    """Generate an HD wallet Account from the supplied master_secret seed, at the given HD derivation
    path, for the specified cryptocurrency.

    """
    acct			= Account(
        crypto		= crypto or 'ETH',
        format		= format,
    ).from_seed(
        seed		= master_secret,
        path		= path,
    )
    return acct


def accounts(
    master_secret: Union[str,bytes],
    crypto: str			= None,  # default 'ETH'
    paths: str			= None,  # default to the crypto's path_default; allow ranges
    format: str			= None,
    allow_unbounded		= True,
):
    """Create accounts for crypto, at the provided paths (allowing ranges), with the optionsal address format. """
    for path in [None] if paths is None else path_sequence( *path_parser(
        paths		= paths,
        allow_unbounded	= allow_unbounded,
    )):
        yield account( master_secret, crypto=crypto, path=path, format=format )


def accountgroups(
    master_secret: bytes,
    cryptopaths: Sequence[Tuple[str,str]],
    allow_unbounded: bool	= True,
) -> Sequence[Sequence[Account]]:
    """Generate the desired cryptocurrency account(s) at each crypto's given path(s).  This is useful
    for generating sequences of groups of wallets for multiple cryptocurrencies, eg. for receiving
    multiple cryptocurrencies for each client.  Since each cryptocurrency uses a different BIP-44 path,
    we have to generate different sequences.

    Supports ranges in each path segment, eg.:

        ('ETH', "m/44'/60'/0'/0/-")	-- generates all accounts for ETH
        ('BTC', "m/44'/0'/0'/0/-")	-- generates all accounts for BTC

        [
          [ "m/44'/60'/0'/0/0", "0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1"],
          [ "m/44'/0'/0'/0/0", "1MAjc529bjmkC1iCXTw2XMHL2zof5StqdQ"]
        ],
        [
          [ "m/44'/60'/0'/0/1", "0x8D342083549C635C0494d3c77567860ee7456963"],
          [ "m/44'/0'/0'/0/1", "1BGwDuVPJeXDG9upaHvVPds5MXwkTjZoav"]
        ],
        ...

    """
    yield from zip( *[
        accounts(
            master_secret	= master_secret,
            paths		= paths,
            crypto		= crypto,
            allow_unbounded	= allow_unbounded,
        )
        for crypto,paths in cryptopaths
    ])


def address(
    master_secret: bytes,
    crypto: str			= None,
    path: str			= None,
    format: str			= None,
):
    """Return the specified cryptocurrency HD account address at path."""
    return account(
        master_secret,
        path		= path,
        crypto		= crypto,
        format		= format,
    ).address


def addresses(
    master_secret: bytes,
    crypto: str	 		= None,  # default 'ETH'
    paths: str			= None,  # default: The crypto's path_default; supports ranges
    format: str			= None,
    allow_unbounded: bool	= True,
):
    """Generate a sequence of cryptocurrency account (path, address, ...)  for all designated
    cryptocurrencies.  Usually a single (<path>, <address>) tuple is desired (different
    cryptocurrencies typically have their own unique path derivations.

    """
    for acct in accounts( master_secret, crypto, paths, format, allow_unbounded=allow_unbounded ):
        yield (acct.crypto, acct.path, acct.address)


def addressgroups(
    master_secret: bytes,
    cryptopaths: Sequence[Tuple[str,str]],
    allow_unbounded: bool	= True,
) -> Sequence[str]:
    """Yields account (<crypto>, <path>, <address>) records for the desired cryptocurrencies at paths.

    """
    yield from zip( *[
        addresses(
            master_secret	= master_secret,
            paths		= paths,
            crypto		= crypto,
            allow_unbounded	= allow_unbounded,
        )
        for crypto,paths in cryptopaths
    ])



import codecs
import contextlib
#import json

import pytest

import shamir_mnemonic

from mnemonic		import Mnemonic

SEED_KNOWN_HEX			= b'87e39270d1d1976e9ade9cc15a084c62'
SEED_KNOWN			= codecs.decode( SEED_KNOWN_HEX, 'hex_codec' )
PASS_KNOWN			= b''

SEED_TREZOR			= b"ABCDEFGHIJKLMNOP"
PASS_TREZOR			= b"TREZOR"

SEED_XMAS_HEX			= b"dd0e2f02b1f6c92a1a265561bc164135"
SEED_XMAS			= codecs.decode( SEED_XMAS_HEX, 'hex_codec' )


class substitute( contextlib.ContextDecorator ):
    """The SLIP-39 standard includes random data in portions of the as share.  Replace the random
    function during testing to get determinism in resultant nmenomics.

    """
    def __init__( self, thing, attribute, value ):
        self.thing		= thing
        self.attribute		= attribute
        self.value		= value
        self.saved		= None

    def __enter__( self ):
        self.saved		= getattr( self.thing, self.attribute )
        setattr( self.thing, self.attribute, self.value )

    def __exit__( self, *exc ):
        setattr( self.thing, self.attribute, self.saved )


def nonrandom_bytes( n ):
    return b'\0' * n


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_share_1_of_3_5():
    mnemonics			= shamir_mnemonic.generate_mnemonics(
        group_threshold	= 1,
        groups		= [(3, 5)],
        master_secret	= SEED_TREZOR,
        passphrase	= PASS_TREZOR,
    )

    #print( json.dumps( mnemonics, indent=4 ))
    assert shamir_mnemonic.combine_mnemonics(mnemonics[0][:3]) \
        == shamir_mnemonic.combine_mnemonics(mnemonics[0][2:])
    assert mnemonics[0][0] == "academic acid academic acne academic academic academic academic academic academic academic academic academic academic academic academic academic carpet making building"
    assert mnemonics[0][1] == "academic acid academic agree artist uncover slavery vocal airport sharp explain violence enlarge minister dragon soul wrist much valid pencil"


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_share_2_of_groups():
    mnemonics		= shamir_mnemonic.generate_mnemonics(
        group_threshold	= 2,
        groups		= [(1, 1), (1, 1), (2, 5), (3, 6)],
        master_secret	= SEED_KNOWN,
        passphrase	= PASS_KNOWN,
    )

    #print( json.dumps( mnemonics, indent=4 ))
    assert len( mnemonics ) == 4
    assert shamir_mnemonic.combine_mnemonics(mnemonics[0] + mnemonics[2][0:2]) \
        == shamir_mnemonic.combine_mnemonics(mnemonics[1] + mnemonics[2][2:4])
    assert mnemonics[2][0] == "academic acid ceramic roster academic photo daisy indicate salary hunting fancy guest taste diploma express usher regret equip install prevent"
    assert mnemonics[2][1] == "academic acid ceramic scared cubic episode metric intend symbolic overall employer course kind human criminal width game duration maiden favorite"


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
@pytest.mark.parametrize("entropy,expected_BIP39,expected_seed,expected_SLIP39", [
    (
        "00000000000000000000000000000000",
        "abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about",
        "c55257c360c07c72029aebc1b53c05ed0362ada38ead3e3e9efa3708e53495531f09a6987599d18264c1e1c92f2cf141630c7a3c4ab7c81b2f001698e7463b04",
        "academic acid acrobat romp academic facility filter scared decision likely luxury acid aquatic campus submit cleanup style repair taste"
        " funding rebuild exotic busy perfect research curly snake exact seafood geology necklace axle fatigue valuable amuse spray pile union"
        " blue switch parking repeat pumps trend cleanup nuclear breathe guitar jacket party home science recover apart render artwork lair ambition market",
    ),
    (
        "7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
        "legal winner thank year wave sausage worth useful legal winner thank yellow",
        "2e8905819b8723fe2c1d161860e5ee1830318dbf49a83bd451cfb8440c28bd6fa457fe1296106559a3c80937a1c1069be3a3a5bd381ee6260e8d9739fce1f607",
        "academic acid acrobat romp acquire home transfer threaten year remove laundry editor midst climate research observe safari deadline staff"
        " twin downtown cause suitable airport total downtown image boundary tracks hospital brother squeeze phrase webcam grief deploy garbage climate"
        " lawsuit headset license aluminum hawk bedroom evidence thumb thorn shrimp finger easy exercise wisdom warn axis timber paper salon harvest luck",
    ),
])
def test_bip39( entropy, expected_BIP39, expected_seed, expected_SLIP39 ):

    # Generate a BIP39 Mnemonic seed.  This is a one-way process; a BIP39 Mnemonic (and its
    # originating Entropy) is *not* recoverable from the Seed, as the derivation function is not
    # reversible.  The process is Entropy --> BIP39 Mnemonic --> Seed
    m = Mnemonic("english")
    mnemonic = m.to_mnemonic( bytes.fromhex( entropy ))
    assert m.check( mnemonic )
    assert mnemonic == expected_BIP39

    seed = Mnemonic.to_seed(mnemonic, passphrase="TREZOR")
    assert seed.hex() == expected_seed

    # Now that we have the desired seed from the BIP39 process, we *can* save the 512-bit Seed via a
    # (large) SLIP39 Mnemonic.

    mnemonics			= shamir_mnemonic.generate_mnemonics(
        group_threshold	= 2,
        groups		= [(1, 1), (1, 1), (2, 5), (3, 6)],
        master_secret	= seed,
        passphrase	= b"",
    )

    # print( json.dumps( mnemonics, indent=4 ))
    assert len( mnemonics ) == 4
    assert shamir_mnemonic.combine_mnemonics(mnemonics[0] + mnemonics[2][0:2]) \
        == shamir_mnemonic.combine_mnemonics(mnemonics[1] + mnemonics[2][2:4]) \
        == seed
    assert all( len( m.split(' ')) == 59 for g in mnemonics for m in g )
    assert any( m == expected_SLIP39     for g in mnemonics for m in g )



from setuptools import setup

import os
import sys
import glob
import fnmatch

HERE				= os.path.dirname( os.path.abspath( __file__ ))

# Must work if setup.py is run in the source distribution context, or from
# within the packaged distribution directory.
__version__			= None
try:
    exec( open( 'slip39/version.py', 'r' ).read() )
except FileNotFoundError:
    exec( open( 'version.py', 'r' ).read() )

console_scripts			= [
    'slip39		= slip39.main:main',
    'slip39-recovery	= slip39.recovery.main:main',
    'slip39-generator	= slip39.generator.main:main',
    'slip39-gui		= slip39.gui.main:main',
]

entry_points			= {
    'console_scripts': 		console_scripts,
}

install_requires		= open( os.path.join( HERE, "requirements.txt" )).readlines()
tests_require			= open( os.path.join( HERE, "requirements-tests.txt" )).readlines()
extras_require			= {
    option: open( os.path.join( HERE, f"requirements-{option}.txt" )).readlines()
    for option in [
        'serial',	# slip39[serial]: Support serial I/O of generated wallet data
        'json',		# slip39[json]:   Support output of encrypted Ethereum JSON wallets
        'gui',		# slip39[gui]:    Support PySimpleGUI/tkinter Graphical UI App
    ]
}

package_dir			= {
    "slip39":			"./slip39",
    "slip39.recovery":		"./slip39/recovery",
    "slip39.generator":		"./slip39/generator",
    "slip39.gui":		"./slip39/gui",
}

long_description_content_type	= 'text/markdown'
long_description		= """\
Creating Ethereum, Bitcoin and other accounts is complex and fraught with potential for loss of funds.

A BIP-39 seed recovery phrase helps, but a *single* lapse in security dooms the account (and all
derived accounts, in fact).  If someone finds your recovery phrase (or you lose it), the accounts
derived from that seed are /gone/.

The SLIP-39 standard allows you to split the seed between 1, 2, or more groups of several mnemonic
recovery phrases.  This is better, but creating such accounts is difficult; presently, only the
Trezor supports these, and they can only be created "manually".  Writing down 5 or more sets of 20
words is difficult, error-prone and time consuming.

The python-slip39 project exists to assist in the safe creation and documentation of Ethereum HD
Wallet seeds and derived accounts, with various SLIP-39 sharing parameters.  It generates the new
random wallet seed, and generates the expected standard Ethereum account(s) (at derivation path
=m/44'/60'/0'/0/0= by default) and Bitcoin accounts (at Bech32 derivation path =m/84'/0'/0'/0/0= by
default), with wallet address and QR codee (compatible with Trezor derivations).  It produces the
required SLIP-39 phrases, and outputs a single PDF containing all the required printable cards to
document the seed (and the specified derived accounts).

On an secure (ideally air-gapped) computer, new seeds can safely be generated and the PDF saved to a
USB drive for printing (or directly printed without the file being saved to disk.).  Presently,
=slip39= can output example ETH, BTC, LTC and DOGE addresses derived from the seed, to illustrate
what accounts are associated with the backed-up seed.  Recovery of the seed to a Trezor is simple,
by entering the mnemonics right on the device.

    $ python3 -m slip39 -v Personal      # or run: slip39 -v Personal
    2022-01-26 13:55:30 slip39           First(1/1): Recover w/ 2 of 4 groups First(1), Second(1), Fam(2/4), Frens(2/6)
    2022-01-26 13:55:30 slip39           1st  1 sister     8 cricket   15 unhappy
    2022-01-26 13:55:30 slip39                2 acid       9 mental    16 ocean
    2022-01-26 13:55:30 slip39                3 acrobat   10 veteran   17 mayor
    2022-01-26 13:55:30 slip39                4 romp      11 phantom   18 promise
    2022-01-26 13:55:30 slip39                5 anxiety   12 grownup   19 wrote
    2022-01-26 13:55:30 slip39                6 laser     13 skunk     20 romp
    2022-01-26 13:55:30 slip39                7 cricket   14 anatomy
    2022-01-26 13:55:30 slip39           Second(1/1): Recover w/ 2 of 4 groups First(1), Second(1), Fam(2/4), Frens(2/6)
    2022-01-26 13:55:30 slip39           1st  1 sister     8 belong    15 spirit
    2022-01-26 13:55:30 slip39                2 acid       9 survive   16 royal
    2022-01-26 13:55:30 slip39                3 beard     10 home      17 often
    2022-01-26 13:55:30 slip39                4 romp      11 herd      18 silver
    2022-01-26 13:55:30 slip39                5 again     12 mountain  19 grocery
    2022-01-26 13:55:30 slip39                6 orbit     13 august    20 antenna
    2022-01-26 13:55:30 slip39                7 very      14 evening
    2022-01-26 13:55:30 slip39           Fam(2/4): Recover w/ 2 of 4 groups First(1), Second(1), Fam(2/4), Frens(2/6)
    2022-01-26 13:55:30 slip39           1st  1 sister     8 rainbow   15 husky
    2022-01-26 13:55:30 slip39                2 acid       9 swing     16 crowd
    2022-01-26 13:55:30 slip39                3 ceramic   10 credit    17 learn
    2022-01-26 13:55:30 slip39                4 roster    11 piece     18 priority
    2022-01-26 13:55:30 slip39                5 already   12 puny      19 hand
    2022-01-26 13:55:30 slip39                6 quiet     13 senior    20 watch
    2022-01-26 13:55:30 slip39                7 erode     14 listen
    2022-01-26 13:55:30 slip39           2nd  1 sister     8 holy      15 revenue
    2022-01-26 13:55:30 slip39                2 acid       9 execute   16 junction
    2022-01-26 13:55:30 slip39                3 ceramic   10 lift      17 elite
    2022-01-26 13:55:30 slip39                4 scared    11 spark     18 flexible
    2022-01-26 13:55:30 slip39                5 domestic  12 yoga      19 inform
    2022-01-26 13:55:30 slip39                6 exact     13 medical   20 predator
    2022-01-26 13:55:30 slip39                7 finger    14 grief
    ...
    2022-01-26 13:55:30 slip39           ETH    m/44'/60'/0'/0/0    : 0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1
    2022-01-26 13:55:30 slip39           BTC    m/84'/0'/0'/0/0     : bc1q6u7qk0tepkxdm8wkhpqzwwy0w8zfls9yvghaxq
    ...
    2022-01-26 13:55:30 slip39           Wrote SLIP39-encoded wallet for 'Personal' to: Personal-2022-01-26+13.55.30-ETH-0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1.pdf

Later, if you need to recover the Ethereum wallet, keep entering SLIP-39 mnemonics until the secret
is recovered (invalid/duplicate mnemonics will be ignored):

    $ python3 -m slip39.recovery -v      # or run: slip39-recovery -v
    Enter 1st SLIP-39 mnemonic: sister acid acrobat romp anxiety laser cricket cricket mental veteran phantom grownup skunk anatomy unhappy ocean mayor promise wrote romp
    2021-12-29 13:24:25 slip39.recovery  Could not recover SLIP-39 master secret with 1 supplied mnemonics: Insufficient number of mnemonic groups. The required number of groups is 2.
    Enter 2nd SLIP-39 mnemonic: a bc
    2021-12-29 13:24:53 slip39.recovery  Could not recover SLIP-39 master secret with 2 supplied mnemonics: Invalid mnemonic word 'a'.
    Enter 3rd SLIP-39 mnemonic: sister acid ceramic roster already quiet erode rainbow swing credit piece puny senior listen husky crowd learn priority hand watch
    2021-12-29 13:24:58 slip39.recovery  Could not recover SLIP-39 master secret with 3 supplied mnemonics: Invalid mnemonic word 'a'.
    Enter 4th SLIP-39 mnemonic: sister acid ceramic scared domestic exact finger holy execute lift spark yoga medical grief revenue junction elite flexible inform predator
    2021-12-29 13:25:14 slip39.recovery  Recovered SLIP-39 secret with 3 (1st, 3rd, 4th) of 4 supplied mnemonics
    2021-12-29 13:25:14 slip39.recovery  Recovered SLIP-39 secret; To re-generate, send it to: python3 -m slip39 --secret -
    32448aabb50cb6b022fdf17d960720df

Finally, regenerate the Ethereum wallet, perhaps including an encrypted JSON wallet file for import
into a software wallet; note that the same Ethereum wallet address 0x8FBC...3bf1 is recovered:

    $ python3 -m slip39 --secret 32448aabb50cb6b022fdf17d960720df --json -
    2022-01-26 14:06:14 slip39           It is recommended to not use '-s|--secret <hex>'; specify '-' to read from input
    2022-01-26 14:06:14 slip39           ETH    m/44'/60'/0'/0/0    : 0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1
    2022-01-26 14:06:14 slip39           BTC    m/84'/0'/0'/0/0     : bc1q6u7qk0tepkxdm8wkhpqzwwy0w8zfls9yvghaxq
    JSON key file password:
    2022-01-26 14:06:21 slip39           Wrote JSON SLIP39's encrypted ETH wallet 0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1 \
                                                derived at m/44'/60'/0'/0/0 to: SLIP39-2022-01-26+14.06.14-ETH-0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1.json
    2022-01-26 14:06:21 slip39           Wrote SLIP39-encoded wallet for '' to: SLIP39-2022-01-26+14.06.14-ETH-0x8FBCe53111817DcE01F9f4C4A6319eA1Ca0c3bf1.pdf


The whole toolchain is suitable for pipelining:

    $ python3 -m slip39 --text --no-card -q \\
        | sort -r \\
        | python3 -m slip39.recovery \\
        | python3 -m slip39 --secret - --no-card -q
    2021-12-28 10:55:17 slip39           ETH m/44'/60'/0'/0/0    : 0x68dD9B59D5dF605f4e9612E8b427Ab31187E2C54
    2021-12-28 10:55:18 slip39.recovery  Recovered SLIP-39 secret with 4 (1st, 2nd, 7th, 8th) of 8 supplied mnemonics
    2021-12-28 10:55:18 slip39           ETH m/44'/60'/0'/0/0    : 0x68dD9B59D5dF605f4e9612E8b427Ab31187E2C54

Here's an example of PDF containing the SLIP-39 recovery mnemonic cards produced:

![slip39 mnemonic cards][slip39-pdf]

[slip39-pdf]: https://github.com/pjkundert/python-slip39/raw/master/images/slip39-pdf.png "slip39 mnemonic cards PDF"
"""

classifiers			= [
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "License :: Other/Proprietary License",
    "Programming Language :: Python :: 3",
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Intended Audience :: Financial and Insurance Industry",
    "Environment :: Console",
    "Topic :: Security :: Cryptography",
    "Topic :: Office/Business :: Financial",
]
project_urls			= {
    "Bug Tracker": "https://github.com/pjkundert/python-slip39/issues",
}

'''
# For py2{app,exe} App Generation.  TODO: Does not work; use PyInstaller instead
mainscript			= "SLIP39.py"

if sys.platform == 'darwin':
    extra_options		= dict(
        setup_requires	= [ 'py2app' ],
        app		= [ mainscript ],
        # Cross-platform applications generally expect sys.argv to
        # be used for opening files.
        # Don't use this with GUI toolkits, the argv
        # emulator causes problems and toolkits generally have
        # hooks for responding to file-open events.
        options		= dict(
            py2app	= dict(
                argv_emulation	= True,
                iconfile	= 'images/SLIP39.icns',
                includes	= 'tkinter',
            ),
        ),
    )
elif sys.platform == 'win32':
    extra_options		= dict(
        setup_requires	= [ 'py2exe' ],
        app		= [ mainscript ],
    )
else:
    extra_options		= dict(
        # Normally unix-like platforms will use "setup.py install"
        # and install the main script as such
        scripts		= [ mainscript ],
    )
'''

setup(
    name			= "slip39",
    version			= __version__,
    install_requires		= install_requires,
    tests_require		= tests_require,
    extras_require		= extras_require,
    packages			= package_dir.keys(),
    package_dir			= package_dir,
    zip_safe			= True,
    entry_points		= entry_points,
    author			= "Perry Kundert",
    author_email		= "perry@dominionrnd.com",
    project_urls		= project_urls,
    description			= "The slip39 module implements SLIP39 recovery for Ethereum accounts",
    long_description		= long_description,
    long_description_content_type = long_description_content_type,
    license			= "Dual License; GPLv3 and Proprietary",
    keywords			= "Ethereum cryptocurrency SLIP39 BIP39 seed recovery",
    url				= "https://github.com/pjkundert/python-slip39",
    classifiers			= classifiers,
    python_requires		= ">=3.9",
    #**extra_options
)



import getpass
import logging
import sys

log_cfg				= {
    "level":	logging.WARNING,
    "datefmt":	'%Y-%m-%d %H:%M:%S',
    #"format":	'%(asctime)s.%(msecs).03d %(threadName)10.10s %(name)-16.16s %(levelname)-8.8s %(funcName)-10.10s %(message)s',
    "format":	'%(asctime)s %(name)-16.16s %(message)s',
}

log_levelmap 			= {
    -2: logging.FATAL,
    -1: logging.ERROR,
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
}


def log_level( adjust ):
    """Return a logging level corresponding to the +'ve/-'ve adjustment"""
    return log_levelmap[
        max(
            min(
                adjust,
                max( log_levelmap.keys() )
            ),
            min( log_levelmap.keys() )
        )
    ]


def ordinal(num):
    ordinal_dict		= {1: "st", 2: "nd", 3: "rd"}
    q, mod			= divmod( num, 10 )
    suffix			= q % 10 != 1 and ordinal_dict.get(mod) or "th"
    return f"{num}{suffix}"


def input_secure( prompt, secret=True, file=None ):
    """When getting secure (optionally secret) input from standard input, we don't want to use getpass, which
    attempts to read from /dev/tty.

    """
    if ( file or sys.stdin ).isatty():
        # From TTY; provide prompts, and do not echo secret input
        if secret:
            return getpass.getpass( prompt, stream=file )
        elif file:
            # Coming from some file; no prompt, read a line from the file source
            return file.readline()
        else:
            return input( prompt )
    else:
        # Not a TTY; don't litter pipeline output with prompts
        if file:
            return file.readline()
        return input()



import argparse
import codecs
import logging

from .api		import random_secret
from .util		import log_cfg, log_level, input_secure
from .layout		import write_pdfs
from .types		import Account
from .defaults		import (   # noqa: F401
    CARD, CARD_SIZES, PAPER,
    BITS, BITS_DEFAULT,
    FILENAME_FORMAT,
    FILENAME_KEYWORDS,
    CRYPTO_PATHS,
)

log				= logging.getLogger( __package__ )


def main( argv=None ):
    ap				= argparse.ArgumentParser(
        description = "Create and output SLIP39 encoded Ethereum wallet(s) to a PDF file.",
        epilog = "" )
    ap.add_argument( '-v', '--verbose', action="count",
                     default=0,
                     help="Display logging information." )
    ap.add_argument( '-q', '--quiet', action="count",
                     default=0,
                     help="Reduce logging output." )
    ap.add_argument( '-o', '--output',
                     default=FILENAME_FORMAT,
                     help=f"Output PDF to file or '-' (stdout); formatting w/ {', '.join( FILENAME_KEYWORDS )} allowed" )
    ap.add_argument( '-t', '--threshold',
                     default=None,
                     help="Number of groups required for recovery (default: half of groups, rounded up)" )
    ap.add_argument( '-g', '--group', action='append',
                     help="A group name[[<require>/]<size>] (default: <size> = 1, <require> = half of <size>, rounded up, eg. 'Frens(3/5)' )." )
    ap.add_argument( '-f', '--format', action='append',
                     default=[],
                     help=f"Specify default crypto address formats: {', '.join( Account.FORMATS )}; default {', '.join( f'{c}:{Account.address_format(c)}' for c in Account.CRYPTOCURRENCIES)}" )
    ap.add_argument( '-c', '--cryptocurrency', action='append',
                     default=[],
                     help=f"A crypto name and optional derivation path ('../<range>/<range>' allowed); defaults: {', '.join( f'{c}:{Account.path_default(c)}' for c in Account.CRYPTOCURRENCIES)}" )
    ap.add_argument( '-j', '--json',
                     default=None,
                     help="Save an encrypted JSON wallet for each Ethereum address w/ this password, '-' reads it from stdin (default: None)" )
    ap.add_argument( '-s', '--secret',
                     default=None,
                     help="Use the supplied 128-, 256- or 512-bit hex value as the secret seed; '-' reads it from stdin (eg. output from slip39.recover)" )
    ap.add_argument( '--bits',
                     default=None,  # Do not enforce default of 128-bit seeds
                     help=f"Ensure that the seed is of the specified bit length; {', '.join( map( str, BITS ))} supported." )
    ap.add_argument( '--passphrase',
                     default=None,
                     help="Encrypt the master secret w/ this passphrase, '-' reads it from stdin (default: None/'')" )
    ap.add_argument( '-C', '--card',
                     default=None,
                     help=f"Card size; {', '.join(CARD_SIZES.keys())} or '(<h>,<w>),<margin>' (default: {CARD})" )
    ap.add_argument( '--paper',
                     default=None,
                     help=f"Paper size (default: {PAPER})" )
    ap.add_argument( '--no-card', dest="card", action='store_false',
                     help="Disable PDF SLIP-39 mnemonic card output" )
    ap.add_argument( '--text', action='store_true',
                     default=None,
                     help="Enable textual SLIP-39 mnemonic output to stdout" )
    ap.add_argument( 'names', nargs="*",
                     help="Account names to produce")
    args			= ap.parse_args( argv )

    # Set up logging; also, handle the degenerate case where logging has *already* been set up (and
    # basicConfig is a NO-OP), by (also) setting the logging level
    log_cfg['level']		= log_level( args.verbose - args.quiet )
    logging.basicConfig( **log_cfg )
    if args.verbose:
        logging.getLogger().setLevel( log_cfg['level'] )

    # If any --format <crypto>:<format> address formats provided
    for cf in args.format:
        try:
            Account.address_format( *cf.split( ':' ) )
        except Exception as exc:
            log.error( f"Invalid address format: {cf}: {exc}" )
            raise

    bits_desired		= int( args.bits ) if args.bits else BITS_DEFAULT

    master_secret		= args.secret
    if master_secret:
        # Master secret seed supplied as hex
        if master_secret == '-':
            master_secret	= input_secure( 'Master secret hex: ', secret=True )
        else:
            log.warning( "It is recommended to not use '-s|--secret <hex>'; specify '-' to read from input" )
        if master_secret.lower().startswith('0x'):
            master_secret	= master_secret[2:]
        master_secret		= codecs.decode( master_secret, 'hex_codec' )
    else:
        # Generate a random secret seed
        master_secret		= random_secret( bits_desired // 8 )
    master_secret_bits		= len( master_secret ) * 8
    if master_secret_bits not in BITS:
        raise ValueError( f"A {master_secret_bits}-bit master secret was supplied; One of {BITS!r} expected" )
    if args.bits and master_secret_bits != bits_desired:  # If a certain seed size specified, enforce
        raise ValueError( f"A {master_secret_bits}-bit master secret was supplied, but {bits_desired} bits was specified" )

    # Optional passphrase
    passphrase			= args.passphrase or ""
    if passphrase == '-':
        passphrase		= input_secure( 'Master seed passphrase: ', secret=True )
    elif passphrase:
        log.warning( "It is recommended to not use '-p|--passphrase <password>'; specify '-' to read from input" )

    try:
        write_pdfs(
            names		= args.names,
            master_secret	= master_secret,
            passphrase		= passphrase,
            group		= args.group,
            threshold		= args.threshold,
            cryptocurrency	= args.cryptocurrency,
            card		= args.card,
            paper		= args.paper,
            filename		= args.output,
            json_pwd		= args.json,
            text		= args.text,
        )
    except Exception as exc:
        log.exception( f"Failed to write PDFs: {exc}" )
        return 1
    return 0



import ast
import io
import os
import json
import logging
import math

from datetime		import datetime
from collections	import namedtuple
from collections.abc	import Callable

from typing		import Dict, List, Tuple, Sequence
import qrcode
from fpdf		import FPDF, FlexTemplate

from .api		import create, enumerate_mnemonic, group_parser
from .util		import input_secure
from .types		import Account
from .defaults		import (
    FONTS, MNEM_ROWS_COLS, CARD, CARD_SIZES, PAPER, PAGE_MARGIN,
    GROUPS, GROUP_THRESHOLD_RATIO,
    FILENAME_FORMAT,
    CRYPTO_PATHS,
)

# Optionally support output of encrypted JSON files
eth_account			= None
try:
    import eth_account
except ImportError:
    pass

log				= logging.getLogger( __package__ )


class Region:
    """Takes authority for a portion of another Region, and places things in it, relative to its
    upper-left and lower-right corners."""
    def __init__( self, name, x1=None, y1=None, x2=None, y2=None, rotate=None ):
        self.name		= name
        self.x1			= x1		# could represent positions, offsets or ratios
        self.y1			= y1
        self.x2			= x2
        self.y2			= y2
        self.rotate		= rotate
        self.regions		= []

    def element( self ):
        "Converts to mm."
        return dict(
            name	= self.name,
            x1		= self.x1 * 25.4,
            y1		= self.y1 * 25.4,
            x2		= self.x2 * 25.4,
            y2		= self.y2 * 25.4,
            rotate	= self.rotate or 0.0
        )

    def __str__( self ):
        return f"({self.name:16}: ({float(self.x1):8}, {float( self.y1):8}) - ({float(self.x2):8} - {float(self.y2):8})"

    def add_region_relative( self, region ):
        region.x1	       += self.x1
        region.y1	       += self.y1
        region.x2	       += self.x2
        region.y2	       += self.y2
        self.regions.append( region )
        return region

    def add_region_proportional( self, region ):
        region.x1		= self.x1 + ( self.x2 - self.x1 ) * ( 0 if region.x1 is None else region.x1 )
        region.y1		= self.y1 + ( self.y2 - self.y1 ) * ( 0 if region.y1 is None else region.y1 )
        region.x2		= self.x1 + ( self.x2 - self.x1 ) * ( 1 if region.x2 is None else region.x2 )
        region.y2		= self.y1 + ( self.y2 - self.y1 ) * ( 1 if region.y2 is None else region.y2 )
        self.regions.append( region )
        return region

    def dimensions( self ):
        "Converts to mm."
        return Coordinate( (self.x2 - self.x1) * 25.4, (self.y2 - self.y1) * 25.4 )

    def elements( self ):
        """Yield a sequence of { 'name': "...", 'x1': #,  ... }"""
        if self.__class__ != Region:
            yield self.element()
        for r in self.regions:
            for d in r.elements():
                yield d


class Text( Region ):
    def __init__( self, *args, font=None, text=None, size=None, size_ratio=None, align=None,
                  foreground=None, background=None, bold=None, italic=None, underline=None,
                  **kwds ):
        self.font		= font
        self.text		= text
        self.size		= size
        self.size_ratio		= size_ratio or 3/4
        self.align		= align
        self.foreground		= foreground
        self.background		= background
        self.bold		= bold
        self.italic		= italic
        self.underline		= underline
        super().__init__( *args, **kwds )

    def element( self ):
        d			= super().element()
        d['type']		= 'T'
        d['font']		= FONTS.get( self.font ) or self.font or FONTS.get( 'sans' )
        line_height		= (self.y2 - self.y1) * 72      # points
        d['size']		= self.size or ( line_height * self.size_ratio )  # No need for int(round(..))
        if self.text is not None:
            d['text']		= self.text
        if self.bold:
            d['bold']		= True
        if self.italic:
            d['italic']		= True
        if self.underline:
            d['underline']	= True
        if self.align is not None:
            d['align']		= self.align
        if self.foreground is not None:
            d['foreground']	= self.foreground
        if self.background is not None:
            d['background']	= self.background
        return d


class Image( Region ):
    def __init__( self, *args, font=None, text=None, size=None, **kwds ):
        self.font		= font
        self.text		= text
        self.size		= None
        super().__init__( *args, **kwds )

    def element( self ):
        d			= super().element()
        d['type']		= 'I'
        return d


class Line( Region ):
    def emit( self, d, *args, **kwds ):
        d['type']		= 'L'
        super().emit( *args, **kwds )


class Box( Region ):
    def element( self ):
        d			= super().element()
        d['type']		= 'B'
        return d


Coordinate			= namedtuple( 'Coordinate', ('x', 'y') )


def layout_card(
    card_size: Coordinate,
    card_margin: int,
    num_mnemonics: int	= 20,
):
    card			= Box( 'card', 0, 0, card_size.x, card_size.y )
    card_interior		= card.add_region_relative(
        Region( 'card-interior', x1=+card_margin, y1=+card_margin, x2=-card_margin, y2=-card_margin )
    )
    for g_n in range( 16 ):
        o = "BB"
        h = "DD"
        f = "FF"
        color			= [
            # Primary
            f"0x{o}{o}{f}",  # Blue
            f"0x{o}{f}{o}",  # Green
            f"0x{f}{o}{o}",  # Red
            # Secondary
            f"0x{o}{f}{f}",  # Cyan,
            f"0x{f}{o}{f}",  # Magenta
            f"0x{f}{f}{o}",  # Yellow
            # Tertiary
            f"0x{o}{h}{f}",  # Ocean
            f"0x{o}{f}{h}",  # Turquoise
            f"0x{f}{o}{h}",  # Red-Magenta
            f"0x{h}{o}{f}",  # Violet
            f"0x{h}{f}{o}",  # Lime
            f"0x{f}{h}{o}",  # Orange
            # Other
            f"0x{o}{h}{h}",  # Light Cyan
            f"0x{h}{o}{h}",  # Light Magenta
            f"0x{h}{h}{o}",  # Light Yellow
            f"0x{h}{h}{h}",  # Light grey,
        ]
        card_interior.add_region_proportional(
            Text( f'card-g{g_n}', x1=1/8, y1=-1/16, x2=7/8, y2=5/16, foreground=int( color[g_n], 16 ), rotate=-45 )
        )
    card_top			= card_interior.add_region_proportional(
        Region( 'card-top', x1=0, y1=0, x2=1, y2=1/4 )
    )
    card_bottom			= card_interior.add_region_proportional(
        Region( 'card-bottom', x1=0, y1=1/4, x2=1, y2=1 )
    )
    card_mnemonics		= card_bottom.add_region_proportional(
        Region( 'card-mnemonics', x1=0, y1=0, x2=13/16, y2=1 )
    )
    card_qr1			= card_bottom.add_region_proportional(
        Image( 'card-qr1', x1=13/16, y1=0, x2=1, y2=1/2 )
    )
    card_qr1.y2			= card_qr1.y1 + (card_qr1.x2 - card_qr1.x1)  # make height same as width

    card_qr2			= card_bottom.add_region_proportional(
        Image( 'card-qr2', x1=13/16, y1=1/2, x2=1, y2=1 )
    )
    card_qr2.y1			= card_qr2.y2 - (card_qr2.x2 - card_qr2.x1)  # make height same as width

    card_top.add_region_proportional(
        Text( 'card-title', x1=0, y1=0, x2=1, y2=40/100, bold=True )
    )
    card_top.add_region_proportional(
        Text( 'card-requires', x1=0, y1=40/100, x2=1, y2=66/100, align='C', italic=True )
    )
    card_top.add_region_proportional(
        Text( 'card-crypto1', x1=0, y1=66/100, x2=1, y2=83/100, align='R' )
    )
    card_top.add_region_proportional(
        Text( 'card-crypto2', x1=0, y1=83/100, x2=1, y2=100/100, align='R' )
    )

    assert num_mnemonics in MNEM_ROWS_COLS, \
        f"Invalid SLIP-39 mnemonic word count: {num_mnemonics}"
    rows,cols		= MNEM_ROWS_COLS[num_mnemonics]
    for r in range( rows ):
        for c in range( cols ):
            card_mnemonics.add_region_proportional(
                Text(
                    f"mnem-{c * rows + r}",
                    x1		= c/cols,
                    y1		= r/rows,
                    x2		= (c+1)/cols,
                    y2		= (r+1)/rows,
                    font	= 'mono',
                    size_ratio	= 9/16,
                )
            )
    return card


def layout_pdf(
        card_dim: Coordinate,                   # mm.
        page_margin_mm: float	= .25 * 25.4,   # mm.
        orientation: str	= 'portrait',
        paper_format: str	= 'Letter'
) -> Tuple[int, FPDF, Callable[[int],[int, Coordinate]]]:
    """Find the ideal orientation for the most cards of the given dimensions.  Returns the number of
    cards per page, the FPDF, and a function useful for laying out templates on the pages of the
    PDF."""
    pdf				= FPDF(
        orientation	= orientation,
        format		= paper_format,
    )
    pdf.set_margin( 0 )

    # FPDF().epw/.eph is *without* page margins, but *with* re-orienting for portrait/landscape
    # Allow 5% bleed over into page margins (to allow for slight error in paper vs. card sizes)
    card_cols			= int(( pdf.epw - page_margin_mm * 2 * 95/100 ) // card_dim.x )
    card_rows			= int(( pdf.eph - page_margin_mm * 2 * 95/100 ) // card_dim.y )
    cards_pp			= card_rows * card_cols

    def page_xy( num ):
        """Returns the page, and the coordinates within that page"""
        page,nth		= divmod( num, cards_pp )
        page_rows,page_cols	= divmod( nth, card_cols )
        return (page, Coordinate( page_margin_mm + page_cols * card_dim.x,
                                  page_margin_mm + page_rows * card_dim.y ))

    return (cards_pp, orientation, pdf, page_xy)


def output_pdf(
    name: str,
    group_threshold: int,
    groups: Dict[str,Tuple[int,List[str]]],
    accounts: Sequence[Sequence[Account]],
    card_format: str		= CARD,   # 'index' or '(<h>,<w>),<margin>'
    paper_format: str		= PAPER,  # 'Letter', ...
) -> Tuple[FPDF, Sequence[Sequence[Account]]]:
    """Produces a PDF, containing a number of cards containing the provided slip39.Details if
    card_format is not False, and pass through the supplied accounts."""

    # Deduce the card size
    try:
        (card_h,card_w),card_margin = CARD_SIZES[card_format.lower()]
    except KeyError:
        (card_h,card_w),card_margin = ast.literal_eval( card_format )
    card_size			= Coordinate( y=card_h, x=card_w )

    # Compute how many cards per page.  Flip page portrait/landscape to match the cards'.  Use the length
    # of the first Group's first Mnemonic to determine the number of mnemonics on each card.
    num_mnemonics		= len( groups[next(iter(groups.keys()))][1][0].split() )
    card			= layout_card(  # converts to mm
        card_size, card_margin, num_mnemonics=num_mnemonics )
    card_dim			= card.dimensions()

    # Find the best PDF and orientation, by max of returned cards_pp (cards per page)
    page_margin_mm		= PAGE_MARGIN * 25.4
    cards_pp,orientation,pdf,page_xy = max(
        layout_pdf( card_dim, page_margin_mm, orientation=orientation, paper_format=paper_format )
        for orientation in ('portrait', 'landscape')
    )
    log.debug( f"Page: {paper_format} {orientation} {pdf.epw:.8}mm w x {pdf.eph:.8}mm h w/ {page_margin_mm}mm margins,"
               f" Card: {card_format} {card_dim.x:.8}mm w x {card_dim.y:.8}mm h == {cards_pp} cards/page" )

    elements			= list( card.elements() )
    if log.isEnabledFor( logging.DEBUG ):
        log.debug( f"Card elements: {json.dumps( elements, indent=4)}" )
    tpl				= FlexTemplate( pdf, list( card.elements() ))

    group_reqs			= list(
        f"{g_nam}({g_of}/{len(g_mns)})" if g_of != len(g_mns) else f"{g_nam}({g_of})"
        for g_nam,(g_of,g_mns) in groups.items() )
    requires			= f"Recover w/ {group_threshold} of {len(group_reqs)} groups {', '.join(group_reqs[:4])}{'...' if len(group_reqs)>4 else ''}"

    # Convert all of the first group's account(s) to an address QR code
    assert accounts and accounts[0], \
        "At least one cryptocurrency account must be supplied"
    qr				= {}
    for i,acct in enumerate( accounts[0] ):
        qrc			= qrcode.QRCode(
            version	= None,
            error_correction = qrcode.constants.ERROR_CORRECT_M,
            box_size	= 10,
            border	= 0
        )
        qrc.add_data( acct.address )
        qrc.make( fit=True )

        qr[i]			= qrc.make_image()
        if log.isEnabledFor( logging.INFO ):
            f			= io.StringIO()
            qrc.print_ascii( out=f )
            f.seek( 0 )
            for line in f:
                log.info( line.strip() )

    card_n			= 0
    page_n			= None
    for g_n,(g_name,(g_of,g_mnems)) in enumerate( groups.items() ):
        for mn_n,mnem in enumerate( g_mnems ):
            p,(offsetx,offsety)	= page_xy( card_n )
            if p != page_n:
                pdf.add_page()
                page_n		= p
            card_n	       += 1

            tpl['card-title']	= f"SLIP39 {g_name}({mn_n+1}/{len(g_mnems)}) for: {name}"
            tpl['card-requires'] = requires
            tpl['card-crypto1']	= f"{accounts[0][0].crypto} {accounts[0][0].path}: {accounts[0][0].address}"
            tpl['card-qr1']	= qr[0].get_image()
            if len( accounts[0] ) > 1:
                tpl['card-crypto2'] = f"{accounts[0][1].crypto} {accounts[0][1].path}: {accounts[0][1].address}"
                tpl['card-qr2']	= qr[1].get_image()
            tpl[f'card-g{g_n}']	= f"{g_name:6.6}..{mn_n+1}" if len(g_name) > 7 else f"{g_name} {mn_n+1}"

            for n,m in enumerate_mnemonic( mnem ).items():
                tpl[f"mnem-{n}"] = m

            tpl.render( offsetx=offsetx, offsety=offsety )

    return pdf,accounts


def write_pdfs(
    names		= None,		# sequence of [ <name>, ... ], or { "<name>": <details> }
    master_secret	= None,
    passphrase		= None,		# UTF-8 encoded passphrase; default is '' (empty)
    group		= None,		# Group specifications, [ "Frens(3/6)", ... ]
    threshold		= None,		# int, or 1/2 of groups by default
    cryptocurrency	= None,		# sequence of [ 'ETH:<path>', ... ] to produce accounts for
    card		= None,		# False outputs no PDF
    paper		= None,
    filename		= None,
    json_pwd		= None,		# If JSON wallet output desired, supply '-' or password
    text		= None,		# Truthy outputs SLIP-39 recover phrases to stdout
):
    """Writes a PDF containing a unique SLIP-39 encoded seed for each of the names specified.

    Returns a { "<filename>": <details>, ... } dictionary of all PDF files written, and each of
    their account details.

    """

    # Convert sequence of group specifications into standard { "<group>": (<needs>, <size>) ... }
    if isinstance( group, dict ):
        groups			= group
    else:
        groups			= dict(
            group_parser( g )
            for g in group or GROUPS
        )
    assert groups and all( isinstance( k, str ) and len( v ) == 2 for k,v in groups.items() ), \
        f"Each group member specification must be a '<group>': (<needs>, <size>), not {type(next(groups.items()))}"

    group_threshold		= int( threshold ) if threshold else math.ceil( len( groups ) * GROUP_THRESHOLD_RATIO )

    cryptopaths			= []
    for crypto in cryptocurrency or CRYPTO_PATHS:
        try:
            if isinstance( crypto, str ):
                crypto,paths	= crypto.split( ':' )   # Find the  <crypto>,<path> tuple
            else:
                crypto,paths	= crypto		# Already a <crypto>,<path> tuple?
        except ValueError:
            crypto,paths	= crypto,None
        cryptopaths.append( (crypto,paths) )

    # If account details not provided in names, generate them.
    if not isinstance( names, dict ):
        assert not master_secret or not names or len( names ) == 1, \
            "Creating multiple account details from the same secret entropy doesn't make sense"
        names			= {
            name: create(
                name		= name,
                group_threshold	= group_threshold,
                groups		= groups,
                master_secret	= master_secret,
                passphrase	= passphrase.encode( 'utf-8' ) if passphrase else b'',
                cryptopaths	= cryptopaths,
            )
            for name in names or [ "SLIP39" ]
        }

    # Generate each desired SLIP-39 wallet file.  Supports --card (the default).
    results			= {}
    for name, details in names.items():
        # Get the first group of the accountgroups in details.accounts.
        accounts		= details.accounts
        assert accounts and accounts[0], \
            "At least one --cryptocurrency must be specified"
        for account in accounts[0]:
            log.warning( f"{account.crypto:6} {account.path:20}: {account.address}" )

        if text:
            # Output the SLIP-39 mnemonics as text:
            #    name: mnemonic
            for g_name,(g_of,g_mnems) in details.groups.items():
                for mnem in g_mnems:
                    print( f"{name}{name and ': ' or ''}{mnem}" )

        # Unless --no-card specified, output a PDF containing the SLIP-39 mnemonic recovery cards
        pdf			= None
        if card is not False:
            pdf,_		= output_pdf(
                *details,
                card_format	= card or CARD,
                paper_format	= paper or PAPER )

        now			= datetime.now()

        pdf_name		= ( filename or FILENAME_FORMAT ).format(
            name	= name,
            date	= datetime.strftime( now, '%Y-%m-%d' ),
            time	= datetime.strftime( now, '%H.%M.%S'),
            crypto	= accounts[0][0].crypto,
            path	= accounts[0][0].path,
            address	= accounts[0][0].address,
        )
        if not pdf_name.lower().endswith( '.pdf' ):
            pdf_name	       += '.pdf'

        if json_pwd:
            # If -j|--json supplied, also emit the encrypted JSON wallet.  This may be a *different*
            # password than the SLIP-39 master secret encryption passphrase.  It will be required to
            # decrypt and use the saved JSON wallet file, eg. to load a software Ethereum wallet.
            assert eth_account, \
                "The optional eth-account package is required to support output of encrypted JSON wallets\n" \
                "    python3 -m pip install eth-account"
            assert any( 'ETH' == crypto for crypto,paths in cryptopaths ), \
                "--json is only valid if '--crypto ETH' wallets are specified"
            if json_pwd == '-':
                json_pwd	= input_secure( 'JSON key file password: ', secret=True )
            else:
                log.warning( "It is recommended to not use '-j|--json <password>'; specify '-' to read from input" )

            for eth in (
                account
                for group in accounts
                for account in group
                if account.crypto == 'ETH'
            ):
                json_str	= json.dumps( eth_account.Account.encrypt( eth.key, json_pwd ), indent=4 )
                json_name	= ( filename or FILENAME_FORMAT ).format(
                    name	= name or "SLIP39",
                    date	= datetime.strftime( now, '%Y-%m-%d' ),
                    time	= datetime.strftime( now, '%H.%M.%S'),
                    crypto	= eth.crypto,
                    path	= eth.path,
                    address	= eth.address,
                )
                if json_name.lower().endswith( '.pdf' ):
                    json_name	= json_name[:-4]
                json_name      += '.json'
                while os.path.exists( json_name ):
                    log.error( "ERROR: Will NOT overwrite {json_name}; adding '.new'!" )
                    json_name  += '.new'
                with open( json_name, 'w' ) as json_f:
                    json_f.write( json_str )
                log.warning( f"Wrote JSON {name or 'SLIP39'}'s encrypted ETH wallet {eth.address} derived at {eth.path} to: {json_name}" )

                if pdf:
                    # Add the encrypted JSON account recovery to the PDF also, if generated.
                    pdf.add_page()
                    margin_mm	= PAGE_MARGIN * 25.4
                    pdf.set_margin( 1.0 * 25.4 )

                    col_width	= pdf.epw - 2 * margin_mm
                    pdf.set_font( FONTS['sans'], size=10 )
                    line_height	= pdf.font_size * 1.2
                    pdf.cell( col_width, line_height, json_name )
                    pdf.ln( line_height )

                    pdf.set_font( FONTS['sans'], size=9 )
                    line_height	= pdf.font_size * 1.1

                    for line in json_str.split( '\n' ):
                        pdf.cell( col_width, line_height, line )
                        pdf.ln( line_height )
                    pdf.image( qrcode.make( json_str ).get_image(), h=min(pdf.eph, pdf.epw)/2, w=min(pdf.eph, pdf.epw)/2 )

        if pdf:
            pdf.output( pdf_name )
            log.warning( f"Wrote SLIP39-encoded wallet for {name!r} to: {pdf_name}" )
        results[pdf_name]	= details

    return results
